#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# utils.pyx
#
# Cython file for utility functions (embedded in thick_impl.pyx).
#------------------------------------------------------------------------------

cdef object _convert_from_json_node(dpiJsonNode *node):
    cdef:
        dpiTimestamp *as_timestamp
        dpiIntervalDS *as_interval
        dpiJsonArray *array
        dpiBytes *as_bytes
        dpiJsonObject *obj
        dict dict_value
        list list_value
        int32_t seconds
        DbType dbtype
        uint32_t i
        str key
    if node.nativeTypeNum == DPI_NATIVE_TYPE_NULL:
        return None
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_NUMBER:
        if node.nativeTypeNum == DPI_NATIVE_TYPE_DOUBLE:
            return node.value.asDouble
        elif node.nativeTypeNum == DPI_NATIVE_TYPE_FLOAT:
            return node.value.asFloat
        as_bytes = &node.value.asBytes
        return PY_TYPE_DECIMAL(as_bytes.ptr[:as_bytes.length].decode())
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_VARCHAR:
        as_bytes = &node.value.asBytes
        return as_bytes.ptr[:as_bytes.length].decode()
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_RAW:
        return node.value.asBytes.ptr[:node.value.asBytes.length]
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_DATE \
            or node.oracleTypeNum == DPI_ORACLE_TYPE_TIMESTAMP:
        as_timestamp = &node.value.asTimestamp
        return cydatetime.datetime_new(as_timestamp.year, as_timestamp.month,
                                       as_timestamp.day, as_timestamp.hour,
                                       as_timestamp.minute,
                                       as_timestamp.second,
                                       as_timestamp.fsecond // 1000, None)
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_BOOLEAN:
        return node.value.asBoolean
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_INTERVAL_DS:
        as_interval = &node.value.asIntervalDS
        seconds = as_interval.hours * 60 * 60 + as_interval.minutes * 60 + \
                as_interval.seconds
        return cydatetime.timedelta_new(as_interval.days, seconds,
                                        as_interval.fseconds // 1000)
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_JSON_OBJECT:
        obj = &node.value.asJsonObject
        dict_value = {}
        for i in range(obj.numFields):
            key = obj.fieldNames[i][:obj.fieldNameLengths[i]].decode()
            dict_value[key] = _convert_from_json_node(&obj.fields[i])
        return dict_value
    elif node.oracleTypeNum == DPI_ORACLE_TYPE_JSON_ARRAY:
        array = &node.value.asJsonArray
        list_value = [None] * array.numElements
        for i in range(array.numElements):
            list_value[i] = _convert_from_json_node(&array.elements[i])
        return list_value
    dbtype = DbType._from_num(node.oracleTypeNum)
    errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED, name=dbtype.name)


cdef int _convert_from_python(object value, DbType dbtype,
                              ThickDbObjectTypeImpl obj_type_impl,
                              dpiDataBuffer *dbvalue, StringBuffer buf,
                              ThickVarImpl var_impl=None,
                              uint32_t pos=0) except -1:
    cdef:
        uint32_t oracle_type = dbtype.num
        ThickDbObjectImpl obj_impl
        dpiTimestamp *timestamp
        ThickLobImpl lob_impl
        int seconds, status
        JsonBuffer json_buf
    if oracle_type == DPI_ORACLE_TYPE_NUMBER:
        if isinstance(value, bool):
            if value:
                buf.set_value("1")
            else:
                buf.set_value("0")
        elif isinstance(value, (int, float, PY_TYPE_DECIMAL)):
            buf.set_value((<str> cpython.PyObject_Str(value)).encode())
        else:
            message = f"expecting number, got {type(value)}"
            raise TypeError(message)
        dbvalue.asBytes.ptr = buf.ptr
        dbvalue.asBytes.length = buf.length
    elif oracle_type == DPI_ORACLE_TYPE_NATIVE_DOUBLE \
            or oracle_type == DPI_ORACLE_TYPE_NATIVE_FLOAT:
        if oracle_type == DPI_ORACLE_TYPE_NATIVE_DOUBLE:
            dbvalue.asDouble = <double> value
        else:
            dbvalue.asFloat = <float> value
    elif oracle_type == DPI_ORACLE_TYPE_VARCHAR \
            or oracle_type == DPI_ORACLE_TYPE_NVARCHAR \
            or oracle_type == DPI_ORACLE_TYPE_CHAR \
            or oracle_type == DPI_ORACLE_TYPE_NCHAR \
            or oracle_type == DPI_ORACLE_TYPE_LONG_VARCHAR:
        buf.set_value(value)
        dbvalue.asBytes.ptr = buf.ptr
        dbvalue.asBytes.length = buf.length
    elif oracle_type == DPI_ORACLE_TYPE_RAW \
            or oracle_type == DPI_ORACLE_TYPE_LONG_RAW:
        buf.set_value(value)
        dbvalue.asBytes.ptr = buf.ptr
        dbvalue.asBytes.length = buf.length
    elif oracle_type == DPI_ORACLE_TYPE_DATE \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP_LTZ \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP_TZ:
        memset(&dbvalue.asTimestamp, 0, sizeof(dbvalue.asTimestamp))
        timestamp = &dbvalue.asTimestamp
        timestamp.year = cydatetime.PyDateTime_GET_YEAR(value)
        timestamp.month = cydatetime.PyDateTime_GET_MONTH(value)
        timestamp.day = cydatetime.PyDateTime_GET_DAY(value)
        if cydatetime.PyDateTime_Check(value):
            timestamp.hour = cydatetime.PyDateTime_DATE_GET_HOUR(value)
            timestamp.minute = cydatetime.PyDateTime_DATE_GET_MINUTE(value)
            timestamp.second = cydatetime.PyDateTime_DATE_GET_SECOND(value)
            timestamp.fsecond = \
                    cydatetime.PyDateTime_DATE_GET_MICROSECOND(value) * 1000
    elif oracle_type == DPI_ORACLE_TYPE_BOOLEAN:
        dbvalue.asBoolean = <bint> value
    elif oracle_type == DPI_ORACLE_TYPE_NATIVE_INT:
        if isinstance(value, bool):
            dbvalue.asInt64 = 1 if value else 0
        else:
            dbvalue.asInt64 = <int64_t> value
    elif oracle_type == DPI_ORACLE_TYPE_INTERVAL_DS:
        seconds = cydatetime.timedelta_seconds(value)
        dbvalue.asIntervalDS.days = cydatetime.timedelta_days(value)
        dbvalue.asIntervalDS.hours = seconds // 3600
        seconds = seconds % 3600
        dbvalue.asIntervalDS.minutes = seconds // 60
        dbvalue.asIntervalDS.seconds = seconds % 60
        dbvalue.asIntervalDS.fseconds = \
                cydatetime.timedelta_microseconds(value) * 1000
    elif oracle_type == DPI_ORACLE_TYPE_OBJECT:
        if not isinstance(value, PY_TYPE_DB_OBJECT):
            raise TypeError("expecting DbObject")
        obj_impl = <ThickDbObjectImpl> value._impl
        if var_impl is not None:
            if dpiVar_setFromObject(var_impl._handle, pos,
                                    obj_impl._handle) < 0:
                _raise_from_odpi()
        else:
            dbvalue.asObject = obj_impl._handle
    elif oracle_type == DPI_ORACLE_TYPE_CLOB \
            or oracle_type == DPI_ORACLE_TYPE_BLOB \
            or oracle_type == DPI_ORACLE_TYPE_NCLOB:
        if isinstance(value, PY_TYPE_LOB):
            lob_impl = value._impl
            if var_impl is not None:
                if dpiVar_setFromLob(var_impl._handle, pos,
                                     lob_impl._handle) < 0:
                    _raise_from_odpi()
            else:
                dbvalue.asLOB = lob_impl._handle
        else:
            buf.set_value(value)
            dbvalue.asBytes.ptr = buf.ptr
            dbvalue.asBytes.length = buf.length
    elif oracle_type == DPI_ORACLE_TYPE_JSON:
        json_buf = JsonBuffer()
        json_buf.from_object(value)
        if dpiJson_setValue(dbvalue.asJson, &json_buf._top_node) < 0:
            _raise_from_odpi()
    else:
        errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED, name=dbtype.name)


cdef object _convert_oci_attr_to_python(uint32_t attr_type,
                                        dpiDataBuffer *value,
                                        uint32_t value_len):
    """
    Convert an OCI attribute value to a Python value.
    """
    if attr_type == PYO_OCI_ATTR_TYPE_STRING:
        if value.asString == NULL:
            return None
        return value.asString[:value_len].decode()
    elif attr_type == PYO_OCI_ATTR_TYPE_BOOLEAN:
        return value.asBoolean
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT8:
        return value.asUint8
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT16:
        return value.asUint16
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT32:
        return value.asUint32
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT64:
        return value.asUint64
    errors._raise_err(errors.ERR_INVALID_OCI_ATTR_TYPE, attr_type=attr_type)


cdef int _convert_python_to_oci_attr(object value, uint32_t attr_type,
                                     StringBuffer str_buf,
                                     dpiDataBuffer *oci_buf,
                                     void **oci_value,
                                     uint32_t *oci_len) except -1:
    """
    Convert a Python value to the format required by an OCI attribute.
    """
    if attr_type == PYO_OCI_ATTR_TYPE_STRING:
        str_buf.set_value(value)
        oci_value[0] = str_buf.ptr
        oci_len[0] = str_buf.length
    elif attr_type == PYO_OCI_ATTR_TYPE_BOOLEAN:
        oci_buf.asBoolean = value
        oci_value[0] = &oci_buf.asBoolean
        oci_len[0] = sizeof(oci_buf.asBoolean)
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT8:
        oci_buf.asUint8 = value
        oci_value[0] = &oci_buf.asUint8
        oci_len[0] = sizeof(oci_buf.asUint8)
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT16:
        oci_buf.asUint16 = value
        oci_value[0] = &oci_buf.asUint16
        oci_len[0] = sizeof(oci_buf.asUint16)
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT32:
        oci_buf.asUint32 = value
        oci_value[0] = &oci_buf.asUint32
        oci_len[0] = sizeof(oci_buf.asUint32)
    elif attr_type == PYO_OCI_ATTR_TYPE_UINT64:
        oci_buf.asUint64 = value
        oci_value[0] = &oci_buf.asUint64
        oci_len[0] = sizeof(oci_buf.asUint64)
    else:
        errors._raise_err(errors.ERR_INVALID_OCI_ATTR_TYPE,
                          attr_type=attr_type)


cdef object _convert_to_python(ThickConnImpl conn_impl, DbType dbtype,
                               ThickDbObjectTypeImpl obj_type_impl,
                               dpiDataBuffer *dbvalue,
                               int preferred_num_type=NUM_TYPE_FLOAT,
                               bint bypass_decode=False):
    cdef:
        uint32_t oracle_type = dbtype.num
        ThickDbObjectImpl obj_impl
        dpiTimestamp *as_timestamp
        dpiJsonNode *json_node
        ThickLobImpl lob_impl
        uint32_t rowid_length
        dpiBytes *as_bytes
        const char *rowid
        int32_t seconds
    if bypass_decode:
        oracle_type = DPI_ORACLE_TYPE_RAW
    if oracle_type == DPI_ORACLE_TYPE_CHAR \
            or oracle_type == DPI_ORACLE_TYPE_NCHAR \
            or oracle_type == DPI_ORACLE_TYPE_VARCHAR \
            or oracle_type == DPI_ORACLE_TYPE_NVARCHAR \
            or oracle_type == DPI_ORACLE_TYPE_LONG_VARCHAR \
            or oracle_type == DPI_ORACLE_TYPE_LONG_NVARCHAR:
        as_bytes = &dbvalue.asBytes
        return as_bytes.ptr[:as_bytes.length].decode()
    elif oracle_type == DPI_ORACLE_TYPE_NUMBER:
        as_bytes = &dbvalue.asBytes
        if preferred_num_type == NUM_TYPE_INT \
                and memchr(as_bytes.ptr, b'.', as_bytes.length) == NULL:
            return int(as_bytes.ptr[:as_bytes.length])
        elif preferred_num_type == NUM_TYPE_DECIMAL:
            return PY_TYPE_DECIMAL(as_bytes.ptr[:as_bytes.length].decode())
        return float(as_bytes.ptr[:as_bytes.length])
    elif oracle_type == DPI_ORACLE_TYPE_RAW \
            or oracle_type == DPI_ORACLE_TYPE_LONG_RAW:
        as_bytes = &dbvalue.asBytes
        return as_bytes.ptr[:as_bytes.length]
    elif oracle_type == DPI_ORACLE_TYPE_DATE \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP_LTZ \
            or oracle_type == DPI_ORACLE_TYPE_TIMESTAMP_TZ:
        as_timestamp = &dbvalue.asTimestamp
        return cydatetime.datetime_new(as_timestamp.year, as_timestamp.month,
                                       as_timestamp.day, as_timestamp.hour,
                                       as_timestamp.minute,
                                       as_timestamp.second,
                                       as_timestamp.fsecond // 1000, None)
    elif oracle_type == DPI_ORACLE_TYPE_BOOLEAN:
        return dbvalue.asBoolean == 1
    elif oracle_type == DPI_ORACLE_TYPE_NATIVE_DOUBLE:
        return dbvalue.asDouble
    elif oracle_type == DPI_ORACLE_TYPE_NATIVE_FLOAT:
        return dbvalue.asFloat
    elif oracle_type == DPI_ORACLE_TYPE_NATIVE_INT:
        return dbvalue.asInt64
    elif oracle_type == DPI_ORACLE_TYPE_ROWID:
        if dpiRowid_getStringValue(dbvalue.asRowid, &rowid, &rowid_length) < 0:
            _raise_from_odpi()
        return rowid[:rowid_length].decode()
    elif oracle_type == DPI_ORACLE_TYPE_CLOB \
            or oracle_type == DPI_ORACLE_TYPE_BLOB \
            or oracle_type == DPI_ORACLE_TYPE_NCLOB \
            or oracle_type == DPI_ORACLE_TYPE_BFILE:
        lob_impl = ThickLobImpl._create(conn_impl, dbtype, dbvalue.asLOB)
        return PY_TYPE_LOB._from_impl(lob_impl)
    elif oracle_type == DPI_ORACLE_TYPE_OBJECT:
        obj_impl = ThickDbObjectImpl.__new__(ThickDbObjectImpl)
        obj_impl.type = obj_type_impl
        if dpiObject_addRef(dbvalue.asObject) < 0:
            _raise_from_odpi()
        obj_impl._handle = dbvalue.asObject
        return PY_TYPE_DB_OBJECT._from_impl(obj_impl)
    elif oracle_type == DPI_ORACLE_TYPE_INTERVAL_DS:
        seconds = dbvalue.asIntervalDS.hours * 60 * 60 + \
                dbvalue.asIntervalDS.minutes * 60 + \
                dbvalue.asIntervalDS.seconds
        return cydatetime.timedelta_new(dbvalue.asIntervalDS.days, seconds,
                                        dbvalue.asIntervalDS.fseconds // 1000)
    elif oracle_type == DPI_ORACLE_TYPE_JSON:
        if dpiJson_getValue(dbvalue.asJson, DPI_JSON_OPT_NUMBER_AS_STRING,
                            &json_node) < 0:
            _raise_from_odpi()
        return _convert_from_json_node(json_node)
    errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED, name=dbtype.name)

cdef uint32_t _get_native_type_num(DbType dbtype):
    cdef uint32_t oracle_type_num = dbtype.num
    if oracle_type_num == DPI_ORACLE_TYPE_NATIVE_FLOAT:
        return DPI_NATIVE_TYPE_FLOAT
    elif oracle_type_num == DPI_ORACLE_TYPE_NATIVE_DOUBLE:
        return DPI_NATIVE_TYPE_DOUBLE
    elif oracle_type_num == DPI_ORACLE_TYPE_BOOLEAN:
        return DPI_NATIVE_TYPE_BOOLEAN
    elif oracle_type_num == DPI_ORACLE_TYPE_CLOB \
            or oracle_type_num == DPI_ORACLE_TYPE_NCLOB \
            or oracle_type_num == DPI_ORACLE_TYPE_BLOB \
            or oracle_type_num == DPI_ORACLE_TYPE_BFILE:
        return DPI_NATIVE_TYPE_LOB
    elif oracle_type_num == DPI_ORACLE_TYPE_DATE \
            or oracle_type_num == DPI_ORACLE_TYPE_TIMESTAMP \
            or oracle_type_num == DPI_ORACLE_TYPE_TIMESTAMP_LTZ \
            or oracle_type_num == DPI_ORACLE_TYPE_TIMESTAMP_TZ:
        return DPI_NATIVE_TYPE_TIMESTAMP
    elif oracle_type_num == DPI_ORACLE_TYPE_STMT:
        return DPI_NATIVE_TYPE_STMT
    elif oracle_type_num == DPI_ORACLE_TYPE_NATIVE_INT:
        return DPI_NATIVE_TYPE_INT64
    elif oracle_type_num == DPI_ORACLE_TYPE_OBJECT:
        return DPI_NATIVE_TYPE_OBJECT
    elif oracle_type_num == DPI_ORACLE_TYPE_ROWID:
        return DPI_NATIVE_TYPE_ROWID
    elif oracle_type_num == DPI_ORACLE_TYPE_INTERVAL_DS:
        return DPI_NATIVE_TYPE_INTERVAL_DS
    elif oracle_type_num == DPI_ORACLE_TYPE_JSON:
        return DPI_NATIVE_TYPE_JSON
    return DPI_NATIVE_TYPE_BYTES


cdef object _create_new_from_info(dpiErrorInfo *error_info):
    """
    Creates a new error object given a dpiErrorInfo structure
    that is already populated with error information.
    """
    cdef bytes msg_bytes = error_info.message[:error_info.messageLength]
    context = "%s: %s" % (error_info.fnName, error_info.action)
    return errors._Error(msg_bytes.decode(), context, code=error_info.code,
                         offset=error_info.offset,
                         isrecoverable=error_info.isRecoverable,
                         iswarning=error_info.isWarning)


cdef int _raise_from_info(dpiErrorInfo *error_info) except -1:
    """
    Raises an exception given a dpiErrorInfo structure that is already
    populated with error information.
    """
    msg_bytes = error_info.message[:error_info.messageLength]
    context = "%s: %s" % (error_info.fnName, error_info.action)
    error = errors._Error(msg_bytes.decode(), context, code=error_info.code,
                          offset=error_info.offset,
                          isrecoverable=bool(error_info.isRecoverable),
                          iswarning=error_info.isWarning)
    exc_type = get_exception_class(error_info.code)
    raise exc_type(error)


cdef int _raise_from_odpi() except -1:
    """
    Raises an exception from ODPI-C, given that an error has been raised by
    ODPI-C (a return code of -1 has been received).
    """
    cdef dpiErrorInfo error_info
    dpiContext_getError(driver_context, &error_info)
    _raise_from_info(&error_info)


def clientversion():
    """
    Returns the version of the Oracle Client library being used as a 5-tuple.
    The five values are the major version, minor version, update number, patch
    number and port update number.
    """
    cdef dpiVersionInfo info
    global client_version
    if client_version is None:
        if driver_context == NULL:
            errors._raise_err(errors.ERR_INIT_ORACLE_CLIENT_NOT_CALLED)
        if dpiContext_getClientVersion(driver_context, &info) < 0:
            _raise_from_odpi()
        client_version = (
            info.versionNum,
            info.releaseNum,
            info.updateNum,
            info.portReleaseNum,
            info.portUpdateNum
        )
    return client_version


def init_oracle_client(lib_dir=None, config_dir=None, error_url=None,
                       driver_name=None):
    """
    Initialize the Oracle Client library. This method is available externally
    in order to be called with parameters that control how the Oracle Client
    library is initialized. If not called earlier, the first usage of the
    Oracle Client library will cause this method to be called internally.
    """
    cdef:
        bytes lib_dir_bytes, config_dir_bytes, driver_name_bytes
        dpiContextCreateParams params
        dpiErrorInfo error_info
        bytes encoding_bytes
    global driver_context_params
    params_tuple = (lib_dir, config_dir, error_url, driver_name)
    if driver_context != NULL:
        if params_tuple != driver_context_params:
            errors._raise_err(errors.ERR_LIBRARY_ALREADY_INITIALIZED)
        return
    with driver_mode.get_manager(requested_thin_mode=False) as mode_mgr:
        memset(&params, 0, sizeof(dpiContextCreateParams))
        encoding_bytes = constants.ENCODING.encode()
        params.defaultEncoding = encoding_bytes
        if config_dir is None:
            config_dir = defaults.config_dir
        if lib_dir is not None:
            lib_dir_bytes = lib_dir.encode()
            params.oracleClientLibDir = lib_dir_bytes
        if config_dir is not None:
            config_dir_bytes = config_dir.encode()
            params.oracleClientConfigDir = config_dir_bytes
        if driver_name is None:
            driver_name = f"{constants.DRIVER_NAME} thk : {VERSION}"
        driver_name_bytes = driver_name.encode()
        params.defaultDriverName = driver_name_bytes
        if error_url is not None:
            error_url_bytes = error_url.encode()
        else:
            error_url_bytes = constants.INSTALLATION_URL.encode()
        params.loadErrorUrl = error_url_bytes
        if dpiContext_createWithParams(DPI_MAJOR_VERSION, DPI_MINOR_VERSION,
                                       &params, &driver_context,
                                       &error_info) < 0:
            _raise_from_info(&error_info)
        driver_context_params = params_tuple


def init_thick_impl(package):
    """
    Initializes globals after the package has been completely initialized. This
    is to avoid circular imports and eliminate the need for global lookups.
    """
    global PY_TYPE_DB_OBJECT, PY_TYPE_LOB
    PY_TYPE_DB_OBJECT = package.DbObject
    PY_TYPE_LOB = package.LOB



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3000 - Module for testing subscriptions
"""

import threading
import unittest

import oracledb
import test_env

class SubscriptionData:

    def __init__(self, num_messages_expected):
        self.condition = threading.Condition()
        self.num_messages_expected = num_messages_expected
        self.num_messages_received = 0

    def _process_message(self, message):
        pass

    def callback_handler(self, message):
        if message.type != oracledb.EVENT_DEREG:
            self._process_message(message)
            self.num_messages_received += 1
        if message.type == oracledb.EVENT_DEREG or \
                self.num_messages_received == self.num_messages_expected:
            with self.condition:
                self.condition.notify()

    def wait_for_messages(self):
        if self.num_messages_received < self.num_messages_expected:
            with self.condition:
                self.condition.wait(10)


class AQSubscriptionData(SubscriptionData):
    pass


class DMLSubscriptionData(SubscriptionData):

    def __init__(self, num_messages_expected):
        super().__init__(num_messages_expected)
        self.table_operations = []
        self.row_operations = []
        self.rowids = []

    def _process_message(self, message):
        table, = message.tables
        self.table_operations.append(table.operation)
        for row in table.rows:
            self.row_operations.append(row.operation)
            self.rowids.append(row.rowid)


@unittest.skipIf(test_env.get_is_thin(),
                 "thin mode doesn't support subscriptions")
class TestCase(test_env.BaseTestCase):

    @unittest.skip("FIXME: threaded mode fails on close of connection?")
    def test_3000_dml_subscription(self):
        "3000 - test subscription for insert, update, delete and truncate"

        # skip if running on the Oracle Cloud, which does not support
        # subscriptions currently
        if self.is_on_oracle_cloud():
            message = "Oracle Cloud does not support subscriptions currently"
            self.skipTest(message)

        # truncate table in order to run test in known state
        self.cursor.execute("truncate table TestTempTable")

        # expected values
        table_operations = [
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_UPDATE,
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_DELETE,
            oracledb.OPCODE_ALTER | oracledb.OPCODE_ALLROWS
        ]
        row_operations = [
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_UPDATE,
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_DELETE
        ]
        rowids = []

        # set up subscription
        data = DMLSubscriptionData(5)
        connection = test_env.get_connection(events=True)
        sub = connection.subscribe(callback=data.callback_handler,
                                   timeout=10, qos=oracledb.SUBSCR_QOS_ROWIDS)
        sub.registerquery("select * from TestTempTable")
        connection.autocommit = True
        cursor = connection.cursor()

        # insert statement
        cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'test')""")
        cursor.execute("select rowid from TestTempTable where IntCol = 1")
        rowids.extend(r for r, in cursor)

        # update statement
        cursor.execute("""
                update TestTempTable set
                    StringCol1 = 'update'
                where IntCol = 1""")
        cursor.execute("select rowid from TestTempTable where IntCol = 1")
        rowids.extend(r for r, in cursor)

        # second insert statement
        cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (2, 'test2')""")
        cursor.execute("select rowid from TestTempTable where IntCol = 2")
        rowids.extend(r for r, in cursor)

        # delete statement
        cursor.execute("delete TestTempTable where IntCol = 2")
        rowids.append(rowids[-1])

        # truncate table
        cursor.execute("truncate table TestTempTable")

        # wait for all messages to be sent
        data.wait_for_messages()

        # verify the correct messages were sent
        self.assertEqual(data.table_operations, table_operations)
        self.assertEqual(data.row_operations, row_operations)
        self.assertEqual(data.rowids, rowids)

        # test string format of subscription object is as expected
        fmt = "<oracledb.Subscription on <oracledb.Connection to %s@%s>>"
        expected = fmt % \
                   (test_env.get_main_user(), test_env.get_connect_string())
        self.assertEqual(str(sub), expected)

    def test_3001_deprecations(self):
        "3001 - test to verify deprecations"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.subscribe,
                               ip_address='www.oracle.in',
                               ipAddress='www.oracle.in')
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.subscribe, grouping_class=1,
                               groupingClass=1)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.subscribe, grouping_value=3,
                               groupingValue=3)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.subscribe, grouping_type=2,
                               groupingType=2)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.subscribe,
                               client_initiated=True, clientInitiated=True)

    @unittest.skip("multiple subscriptions cannot be created simultaneously")
    def test_3002_aq_subscription(self):
        "3002 - test subscription for AQ"

        # create queue and clear it of all messages
        queue = self.connection.queue("TEST_RAW_QUEUE")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        while queue.deqone():
            pass
        self.connection.commit()

        # set up subscription
        data = AQSubscriptionData(1)
        connection = test_env.get_connection(events=True)
        sub = connection.subscribe(namespace=oracledb.SUBSCR_NAMESPACE_AQ,
                                   name=queue.name, timeout=10,
                                   callback=data.callback_handler)

        # enqueue a message
        queue.enqone(self.connection.msgproperties(payload="Some data"))
        self.connection.commit()

        # wait for all messages to be sent
        data.wait_for_messages()

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# var.pyx
#
# Cython file defining the thick Variable implementation class (embedded in
# thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThickVarImpl(BaseVarImpl):
    cdef:
        dpiVar *_handle
        dpiData *_data
        StringBuffer _buf
        uint32_t _native_type_num
        bint _get_returned_data
        object _conn

    def __dealloc__(self):
        if self._handle != NULL:
            dpiVar_release(self._handle)

    cdef int _bind(self, object conn, BaseCursorImpl cursor_impl,
                   uint32_t num_execs, object name, uint32_t pos) except -1:
        cdef:
            ThickCursorImpl thick_cursor_impl = <ThickCursorImpl> cursor_impl
            uint32_t name_length, i
            dpiDataBuffer *dbvalue
            const char *name_ptr
            bytes name_bytes
        if self._native_type_num == DPI_NATIVE_TYPE_STMT:
            for i in range(self.num_elements):
                if self._data[i].isNull:
                    continue
                dbvalue = &self._data[i].value
                if dbvalue.asStmt == thick_cursor_impl._handle:
                    errors._raise_err(errors.ERR_SELF_BIND_NOT_SUPPORTED)
        if name is not None:
            name_bytes = name.encode()
            name_ptr = name_bytes
            name_length = <uint32_t> len(name_bytes)
            if dpiStmt_bindByName(thick_cursor_impl._handle, name_ptr,
                                  name_length, self._handle) < 0:
                _raise_from_odpi()
        else:
            if dpiStmt_bindByPos(thick_cursor_impl._handle, pos,
                                 self._handle) < 0:
                _raise_from_odpi()
        if thick_cursor_impl._stmt_info.isReturning and not self._is_value_set:
            self._get_returned_data = True
        self._is_value_set = False

    cdef int _create_handle(self) except -1:
        cdef:
             ThickConnImpl conn_impl = self._conn_impl
             dpiObjectType *obj_type_handle = NULL
             ThickDbObjectTypeImpl obj_type_impl
        if self._handle != NULL:
             dpiVar_release(self._handle)
             self._handle = NULL
        if self.objtype is not None:
             obj_type_impl = <ThickDbObjectTypeImpl> self.objtype
             obj_type_handle = obj_type_impl._handle
        self._native_type_num = _get_native_type_num(self.dbtype)
        if dpiConn_newVar(conn_impl._handle, self.dbtype.num,
                          self._native_type_num, self.num_elements, self.size,
                          0, self.is_array, obj_type_handle, &self._handle,
                          &self._data) < 0:
             _raise_from_odpi()

    cdef int _finalize_init(self) except -1:
        """
        Internal method that finalizes initialization of the variable.
        """
        BaseVarImpl._finalize_init(self)
        self._create_handle()

    cdef list _get_array_value(self):
        """
        Internal method to return the value of the array.
        """
        cdef uint32_t i
        return [self._get_scalar_value(i) \
                for i in range(self.num_elements_in_array)]

    cdef object _get_cursor_value(self, dpiDataBuffer *dbvalue):
        cdef:
            ThickCursorImpl cursor_impl
            object cursor
        cursor = self._conn.cursor()
        cursor_impl = <ThickCursorImpl> cursor._impl
        if dpiStmt_addRef(dbvalue.asStmt) < 0:
            _raise_from_odpi()
        cursor_impl._handle = dbvalue.asStmt
        cursor_impl._fixup_ref_cursor = True
        return cursor

    cdef object _get_scalar_value(self, uint32_t pos):
        """
        Internal method to return the value of the variable at the given
        position.
        """
        cdef:
            uint32_t num_returned_rows
            dpiData *returned_data
        if self._get_returned_data:
            if dpiVar_getReturnedData(self._handle, pos, &num_returned_rows,
                                      &returned_data) < 0:
                _raise_from_odpi()
            return self._transform_array_to_python(num_returned_rows,
                                                   returned_data)
        return self._transform_element_to_python(pos, self._data)

    cdef int _on_reset_bind(self, uint32_t num_rows) except -1:
        """
        Called when the bind variable is being reset, just prior to performing
        a bind operation.
        """
        cdef:
            dpiStmtInfo stmt_info
            uint32_t i
        BaseVarImpl._on_reset_bind(self, num_rows)
        if self._native_type_num == DPI_NATIVE_TYPE_STMT:
            for i in range(self.num_elements):
                if self._data[i].isNull:
                    continue
                if dpiStmt_getInfo(self._data[i].value.asStmt, &stmt_info) < 0:
                    self._create_handle()
                    break

    cdef int _resize(self, uint32_t new_size) except -1:
        """
        Resize the variable to the new size provided.
        """
        cdef:
            dpiVar *orig_handle = NULL
            uint32_t num_elements, i
            dpiData *source_data
            dpiData *orig_data
        BaseVarImpl._resize(self, new_size)
        orig_data = self._data
        orig_handle = self._handle
        self._handle = NULL
        try:
            self._create_handle()
            if self.is_array:
                if dpiVar_getNumElementsInArray(orig_handle,
                                                &num_elements) < 0:
                    _raise_from_odpi()
                if dpiVar_setNumElementsInArray(self._handle,
                                                num_elements) < 0:
                    _raise_from_odpi()
            for i in range(self.num_elements):
                source_data = &orig_data[i]
                if source_data.isNull:
                    continue
                if dpiVar_setFromBytes(self._handle, i,
                        source_data.value.asBytes.ptr,
                        source_data.value.asBytes.length) < 0:
                    _raise_from_odpi()
        finally:
            dpiVar_release(orig_handle)

    cdef int _set_cursor_value(self, object cursor, uint32_t pos) except -1:
        cdef:
            ThickCursorImpl cursor_impl = cursor._impl
            dpiData *data

        # if the cursor already has a handle, use it directly
        if cursor_impl._handle != NULL:
            if dpiVar_setFromStmt(self._handle, pos, cursor_impl._handle) < 0:
                _raise_from_odpi()

        # otherwise, make use of the statement handle allocated by the variable
        else:
            data = &self._data[pos]
            if dpiStmt_addRef(data.value.asStmt) < 0:
                _raise_from_odpi()
            cursor_impl._handle = data.value.asStmt
        cursor_impl._fixup_ref_cursor = True
        cursor.statement = None

    cdef int _set_num_elements_in_array(self, uint32_t num_elements) except -1:
        """
        Sets the number of elements in the array.
        """
        BaseVarImpl._set_num_elements_in_array(self, num_elements)
        if dpiVar_setNumElementsInArray(self._handle, num_elements) < 0:
            _raise_from_odpi()

    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1:
        """
        Set the value of the variable at the given position. At this point it
        is assumed that all checks have been performed!
        """
        cdef:
            dpiDataBuffer temp_dbvalue
            dpiDataBuffer *dbvalue
            dpiBytes *as_bytes
            bint needs_set
            dpiData *data
        data = &self._data[pos]
        data.isNull = (value is None)
        if not data.isNull:
            if self._native_type_num == DPI_NATIVE_TYPE_STMT:
                self._set_cursor_value(value, pos)
            else:
                needs_set = self._native_type_num == DPI_NATIVE_TYPE_BYTES \
                        or (self._native_type_num == DPI_NATIVE_TYPE_LOB and \
                                not isinstance(value, PY_TYPE_LOB))
                if needs_set:
                    dbvalue = &temp_dbvalue
                else:
                    dbvalue = &data.value
                _convert_from_python(value, self.dbtype, self.objtype, dbvalue,
                                     self._buf, self, pos)
                if needs_set:
                    as_bytes = &dbvalue.asBytes
                    if dpiVar_setFromBytes(self._handle, pos, as_bytes.ptr,
                                           as_bytes.length) < 0:
                        _raise_from_odpi()

    cdef object _transform_array_to_python(self, uint32_t num_elements,
                                           dpiData *data):
        """
        Transforms an array from ODPI-C to a Python list.
        """
        cdef:
            object element_value
            list return_value
            uint32_t i
        return_value = cpython.PyList_New(num_elements)
        for i in range(num_elements):
            element_value = self._transform_element_to_python(i, data)
            cpython.Py_INCREF(element_value)
            cpython.PyList_SET_ITEM(return_value, i, element_value)
        return return_value

    cdef object _transform_element_to_python(self, uint32_t pos,
                                             dpiData *data):
        """
        Transforms a single element from the value supplied by ODPI-C to its
        equivalent Python value.
        """
        cdef object value
        data = &data[pos]
        if not data.isNull:
            if self._native_type_num == DPI_NATIVE_TYPE_STMT:
                return self._get_cursor_value(&data.value)
            value = _convert_to_python(self._conn_impl, self.dbtype,
                                       self.objtype, &data.value,
                                       self._preferred_num_type,
                                       self.bypass_decode)
            if self.outconverter is not None:
                value = self.outconverter(value)
            return value



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# soda.pyx
#
# Cython file defining the base SODA implementation classes (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseSodaDbImpl:

    @utils.CheckImpls("creating a SODA collection")
    def create_collection(self, str name, str metadata, bint map_mode):
        pass

    @utils.CheckImpls("creating a SODA document")
    def create_document(self, bytes content, str key, str media_type):
        pass

    @utils.CheckImpls("getting a list of SODA collection names")
    def get_collection_names(self, str start_name, uint32_t limit):
        pass

    @utils.CheckImpls("opening a SODA collection")
    def open_collection(self, str name):
        pass


cdef class BaseSodaCollImpl:

    @utils.CheckImpls("creating an index on a SODA collection")
    def create_index(self, str spec):
        pass

    @utils.CheckImpls("dropping a SODA collection")
    def drop(self):
        pass

    @utils.CheckImpls("dropping an index on a SODA collection")
    def drop_index(self, str name, bint force):
        pass

    @utils.CheckImpls("getting the count of documents in a SODA collection")
    def get_count(self, object op):
        pass

    @utils.CheckImpls("getting a cursor for documents in a SODA collection")
    def get_cursor(self, object op):
        pass

    @utils.CheckImpls("getting the data guide for a SODA collection")
    def get_data_guide(self):
        pass

    @utils.CheckImpls("getting the metadata of a SODA collection")
    def get_metadata(self):
        pass

    @utils.CheckImpls("getting a document from a SODA collection")
    def get_one(self, object op):
        pass

    @utils.CheckImpls("inserting multiple documents into a SODA collection")
    def insert_many(self, list documents, str hint, bint return_docs):
        pass

    @utils.CheckImpls("inserting a single document into a SODA collection")
    def insert_one(self, BaseSodaDocImpl doc, str hint, bint return_doc):
        pass

    @utils.CheckImpls("removing documents from a SODA collection")
    def remove(self, object op):
        pass

    @utils.CheckImpls("replacing a document in a SODA collection")
    def replace_one(self, BaseSodaDocImpl doc_impl, bint return_doc):
        pass

    @utils.CheckImpls("saving a document in a SODA collection")
    def save(self, BaseSodaDocImpl doc, str hint, bint return_doc):
        pass

    @utils.CheckImpls("truncating a SODA collection")
    def truncate(self):
        pass


cdef class BaseSodaDocImpl:

    @utils.CheckImpls("getting the content of a SODA document")
    def get_content(self):
        pass

    @utils.CheckImpls("getting the created on date of a SODA document")
    def get_created_on(self):
        pass

    @utils.CheckImpls("getting the key of a SODA document")
    def get_key(self):
        pass

    @utils.CheckImpls("getting the last modified date of a SODA document")
    def get_last_modified(self):
        pass

    @utils.CheckImpls("getting the media type of a SODA document")
    def get_media_type(self):
        pass

    @utils.CheckImpls("getting the version of a SODA document")
    def get_version(self):
        pass


cdef class BaseSodaDocCursorImpl:

    @utils.CheckImpls("closing a SODA document cursor")
    def close(self):
        pass

    @utils.CheckImpls("getting the next document from a SODA document cursor")
    def get_next_doc(self):
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2100 - Module for testing NCHAR variables
"""

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            unicode_col = "Unicode \u3042 %d" % i
            fixed_char_col = ("Fixed Unicode %d" % i).ljust(40)
            if i % 2:
                nullable_col = "Nullable %d" % i
            else:
                nullable_col = None
            data_tuple = (i, unicode_col, fixed_char_col, nullable_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_2100_unicode_length(self):
        "2100 - test value length"
        return_value = self.cursor.var(int)
        self.cursor.execute("""
                begin
                    :retval := LENGTH(:value);
                end;""",
                value="InVal \u3042",
                retval=return_value)
        self.assertEqual(return_value.getvalue(), 7)

    def test_2101_bind_unicode(self):
        "2101 - test binding in a unicode"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute("""
                select * from TestUnicodes
                where UnicodeCol = :value""",
                value="Unicode \u3042 5")
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_2102_bind_different_var(self):
        "2102 - test binding a different variable on second execution"
        retval_1 = self.cursor.var(oracledb.DB_TYPE_NVARCHAR, 30)
        retval_2 = self.cursor.var(oracledb.DB_TYPE_NVARCHAR, 30)
        self.cursor.execute(r"begin :retval := unistr('Called \3042'); end;",
                            retval=retval_1)
        self.assertEqual(retval_1.getvalue(), "Called \u3042")
        self.cursor.execute("begin :retval := 'Called'; end;", retval=retval_2)
        self.assertEqual(retval_2.getvalue(), "Called")

    def test_2103_bind_unicode_after_number(self):
        "2103 - test binding in a string after setting input sizes to a number"
        unicode_val = self.cursor.var(oracledb.DB_TYPE_NVARCHAR)
        unicode_val.setvalue(0, "Unicode \u3042 6")
        self.cursor.setinputsizes(value=oracledb.NUMBER)
        self.cursor.execute("""
                select * from TestUnicodes
                where UnicodeCol = :value""",
                value=unicode_val)
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[6]])

    def test_2104_bind_unicode_array_direct(self):
        "2104 - test binding in a unicode array"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = [r[1] for r in self.raw_data]
        array_var = self.cursor.arrayvar(oracledb.DB_TYPE_NVARCHAR, array)
        statement = """
                begin
                    :retval := pkg_TestUnicodeArrays.TestInArrays(
                        :integer_value, :array);
                end;"""
        self.cursor.execute(statement, retval=return_value, integer_value=5,
                            array=array_var)
        self.assertEqual(return_value.getvalue(), 116)
        array = ["Unicode - \u3042 %d" % i for i in range(15)]
        array_var = self.cursor.arrayvar(oracledb.DB_TYPE_NVARCHAR, array)
        self.cursor.execute(statement, integer_value=8, array=array_var)
        self.assertEqual(return_value.getvalue(), 208)

    def test_2105_bind_unicode_array_by_sizes(self):
        "2105 - test binding in a unicode array (with setinputsizes)"
        return_value = self.cursor.var(oracledb.NUMBER)
        self.cursor.setinputsizes(array = [oracledb.DB_TYPE_NVARCHAR, 10])
        array = [r[1] for r in self.raw_data]
        self.cursor.execute("""
                begin
                    :retval := pkg_TestUnicodeArrays.TestInArrays(:integer_value,
                        :array);
                end;""",
                retval=return_value,
                integer_value=6,
                array=array)
        self.assertEqual(return_value.getvalue(), 117)

    def test_2106_bind_unicode_array_by_var(self):
        "2106 - test binding in a unicode array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.DB_TYPE_NVARCHAR, 10, 20)
        array.setvalue(0, [r[1] for r in self.raw_data])
        self.cursor.execute("""
                begin
                    :retval := pkg_TestUnicodeArrays.TestInArrays(:integer_value,
                        :array);
                end;""",
                retval=return_value,
                integer_value=7,
                array=array)
        self.assertEqual(return_value.getvalue(), 118)

    def test_2107_bind_in_out_unicode_array_by_var(self):
        "2107 - test binding in/out a unicode array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DB_TYPE_NVARCHAR, 10, 100)
        original_data = [r[1] for r in self.raw_data]
        fmt = "Converted element \u3042 # %d originally had length %d"
        expected_data = [fmt % (i, len(original_data[i - 1])) \
                         for i in range(1, 6)] + original_data[5:]
        array.setvalue(0, original_data)
        self.cursor.execute("""
                begin
                    pkg_TestUnicodeArrays.TestInOutArrays(:numElems, :array);
                end;""",
                numElems = 5,
                array = array)
        self.assertEqual(array.getvalue(), expected_data)

    def test_2108_bind_out_unicode_array_by_var(self):
        "2108 - test binding out a unicode array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DB_TYPE_NVARCHAR, 6, 100)
        fmt = "Test out element \u3042 # %d"
        expected_data = [fmt % i for i in range(1, 7)]
        self.cursor.execute("""
                begin
                    pkg_TestUnicodeArrays.TestOutArrays(:numElems, :array);
                end;""",
                numElems = 6,
                array = array)
        self.assertEqual(array.getvalue(), expected_data)

    def test_2109_bind_null(self):
        "2109 - test binding in a null"
        self.cursor.execute("""
                select * from TestUnicodes
                where UnicodeCol = :value""",
                value = None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2110_bind_out_set_input_sizes_by_type(self):
        "2110 - test binding out with set input sizes defined (by type)"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute(r"""
                begin
                    :value := unistr('TSI \3042');
                end;""")
        self.assertEqual(bind_vars["value"].getvalue(), "TSI \u3042")

    def test_2111_bind_in_out_set_input_sizes_by_type(self):
        "2111 - test binding in/out with set input sizes defined (by type)"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute(r"""
                begin
                    :value := :value || unistr(' TSI \3042');
                end;""",
                value = "InVal \u3041")
        self.assertEqual(bind_vars["value"].getvalue(),
                         "InVal \u3041 TSI \u3042")

    def test_2112_bind_out_var(self):
        "2112 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute(r"""
                begin
                    :value := unistr('TSI (VAR) \3042');
                end;""",
                value=var)
        self.assertEqual(var.getvalue(), "TSI (VAR) \u3042")

    def test_2113_bind_in_out_var_direct_set(self):
        "2113 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_NVARCHAR)
        var.setvalue(0, "InVal \u3041")
        self.cursor.execute(r"""
                begin
                    :value := :value || unistr(' TSI (VAR) \3042');
                end;""",
                value = var)
        self.assertEqual(var.getvalue(), "InVal \u3041 TSI (VAR) \u3042")

    def test_2114_cursor_description(self):
        "2114 - test cursor description is accurate"
        self.cursor.execute("select * from TestUnicodes")
        varchar_ratio, nvarchar_ratio = test_env.get_charset_ratios()
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('UNICODECOL', oracledb.DB_TYPE_NVARCHAR, 20, 20 * nvarchar_ratio,
                    None, None, False),
            ('FIXEDUNICODECOL', oracledb.DB_TYPE_NCHAR, 40,
                    40 * nvarchar_ratio, None, None, False),
            ('NULLABLECOL', oracledb.DB_TYPE_NVARCHAR, 50, 50 * nvarchar_ratio,
                    None, None, True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_2115_fetchall(self):
        "2115 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestUnicodes order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2116_fetchmany(self):
        "2116 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestUnicodes order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_2117_fetchone(self):
        "2117 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestUnicodes
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# var.pyx
#
# Cython file defining the variable implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThinVarImpl(BaseVarImpl):
    cdef:
        list _values
        object _conv_func
        object _last_raw_value

    cdef int _bind(self, object conn, BaseCursorImpl cursor_impl,
                   uint32_t num_execs, object name, uint32_t pos) except -1:
        cdef:
            ThinCursorImpl thin_cursor_impl = <ThinCursorImpl> cursor_impl
            Statement stmt = thin_cursor_impl._statement
            object bind_info_dict = stmt._bind_info_dict
            list bind_info_list = stmt._bind_info_list
            ssize_t idx, num_binds, num_vars
            BindInfo bind_info
            str normalized_name
            object value, lob

        # for PL/SQL blocks, if the size of a string or bytes object exceeds
        # 32,767 bytes it must be converted to a BLOB/CLOB; and out converter
        # needs to be established as well to return the string in the way that
        # the user expects to get it
        if stmt._is_plsql and self.size > 32767:
            if self.dbtype._ora_type_num == TNS_DATA_TYPE_RAW \
                    or self.dbtype._ora_type_num == TNS_DATA_TYPE_LONG_RAW:
                self.dbtype = DB_TYPE_BLOB
            elif self.dbtype._csfrm == TNS_CS_NCHAR:
                self.dbtype = DB_TYPE_NCLOB
            else:
                self.dbtype = DB_TYPE_CLOB
            orig_converter = self.outconverter
            def converter(v):
                v = v.read()
                if orig_converter is not None:
                    v = orig_converter(v)
                return v
            self.outconverter = converter

        # for variables containing LOBs, create temporary LOBs, if needed
        if self.dbtype._ora_type_num == TNS_DATA_TYPE_CLOB \
                or self.dbtype._ora_type_num == TNS_DATA_TYPE_BLOB:
            for idx, value in enumerate(self._values):
                if value is not None and not isinstance(value, PY_TYPE_LOB):
                    lob = conn.createlob(self.dbtype)
                    if value:
                        lob.write(value)
                    self._values[idx] = lob

        # bind by name
        if name is not None:
            if name.startswith('"') and name.endswith('"'):
                normalized_name = name[1:-1]
            else:
                normalized_name = name.upper()
            if normalized_name.startswith(":"):
                normalized_name = normalized_name[1:]
            if normalized_name not in bind_info_dict:
                errors._raise_err(errors.ERR_INVALID_BIND_NAME, name=name)
            for bind_info in bind_info_dict[normalized_name]:
                stmt._set_var(bind_info, self, thin_cursor_impl)

        # bind by position
        else:
            num_binds = len(bind_info_list)
            num_vars = len(cursor_impl.bind_vars)
            if num_binds != num_vars:
                errors._raise_err(errors.ERR_WRONG_NUMBER_OF_POSITIONAL_BINDS,
                                  expected_num=num_binds, actual_num=num_vars)
            bind_info = bind_info_list[pos - 1]
            stmt._set_var(bind_info, self, thin_cursor_impl)

    cdef int _finalize_init(self) except -1:
        """
        Internal method that finalizes initialization of the variable.
        """
        BaseVarImpl._finalize_init(self)
        self._values = [None] * self.num_elements

    cdef list _get_array_value(self):
        """
        Internal method to return the value of the array.
        """
        return self._values[:self.num_elements_in_array]

    cdef object _get_scalar_value(self, uint32_t pos):
        """
        Internal method to return the value of the variable at the given
        position.
        """
        return self._values[pos]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1:
        """
        Set the value of the variable at the given position. At this point it
        is assumed that all checks have been performed!
        """
        self._values[pos] = value



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4800 - Module for testing timestamp with local time zone variables
"""

import datetime

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        base_date = datetime.datetime(2022, 6, 2)
        for i in range(1, 11):
            if i % 4 == 0:
                tz_hours = i
            elif i % 2 == 0:
                tz_hours = i + 0.5
            else:
                tz_hours = -(i + 0.5)
            tz_offset = datetime.timedelta(hours=tz_hours)
            microseconds = int(str(i * 50).ljust(6, "0"))
            offset = datetime.timedelta(days=i, seconds=i * 2,
                                        microseconds=microseconds)
            col = base_date + tz_offset + offset
            if i % 2:
                tz_offset = datetime.timedelta(hours=6)
                microseconds = int(str(i * 125).ljust(6, "0"))
                offset = datetime.timedelta(days=i + 1,
                                            seconds=i * 3,
                                            microseconds=microseconds)
                nullable_col = base_date + offset
            else:
                nullable_col = None
            data_tuple = (i, col, nullable_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_4800_bind_timestamp(self):
        "4800 - test binding in a timestamp"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                select * from TestTimestampLTZs
                where TimestampLTZCol = :value""",
                value=datetime.datetime(2022, 6, 6, 18, 30, 10, 250000))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_4801_bind_null(self):
        "4801 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                select * from TestTimestampLTZs
                where TimestampLTZCol = :value""",
                value=None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_4802_bind_out_set_input_sizes(self):
        "4802 - test binding out with set input sizes defined"
        bv = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                begin
                    :value := to_timestamp('20220603', 'YYYYMMDD');
                end;""")
        self.assertEqual(bv["value"].getvalue(), datetime.datetime(2022, 6, 3))

    def test_4803_bind_in_out_set_input_sizes(self):
        "4803 - test binding in/out with set input sizes defined"
        bv = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value=datetime.datetime(2022, 5, 10, 12, 0, 0))
        self.assertEqual(bv["value"].getvalue(),
                         datetime.datetime(2022, 5, 15, 18, 0, 0))

    def test_4804_bind_out_var(self):
        "4804 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                begin
                    :value := to_date('20220601 15:38:12',
                        'YYYYMMDD HH24:MI:SS');
                end;""",
                value=var)
        self.assertEqual(var.getvalue(),
                         datetime.datetime(2022, 6, 1, 15, 38, 12))

    def test_4805_bind_in_out_var_direct_set(self):
        "4805 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP_LTZ)
        var.setvalue(0, datetime.datetime(2022, 5, 30, 6, 0, 0))
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value = var)
        self.assertEqual(var.getvalue(),
                         datetime.datetime(2022, 6, 4, 12, 0, 0))

    def test_4806_cursor_description(self):
        "4806 - test cursor description is accurate"
        self.cursor.execute("select * from TestTimestampLTZs")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('TIMESTAMPLTZCOL', oracledb.DB_TYPE_TIMESTAMP_LTZ, 23, None, 0, 6,
                    False),
            ('NULLABLECOL', oracledb.DB_TYPE_TIMESTAMP_LTZ, 23, None, 0, 6,
                    True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_4807_fetchall(self):
        "4807 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestTimestampLTZs order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_4808_fetchmany(self):
        "4808 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestTimestampLTZs order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_4809_fetchone(self):
        "4809 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestTimestampLTZs
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

    def test_4810_bind_timestamp_with_zero_fseconds(self):
        "4810 - test binding a timestamp with zero fractional seconds"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP_LTZ)
        self.cursor.execute("""
                select *
                from TestTimestampLTZs
                where trunc(TimestampLTZCol) = :value""",
                value=datetime.datetime(2022, 6, 12))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[10]])

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1600 - Module for testing DML returning clauses
"""

import datetime
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def test_1600_insert(self):
        "1600 - test insert (single row) with DML returning"
        self.cursor.execute("truncate table TestTempTable")
        int_val = 5
        str_val = "A test string"
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var""",
                int_val=int_val,
                str_val=str_val,
                int_var=int_var,
                str_var=str_var)
        self.assertEqual(int_var.values, [[int_val]])
        self.assertEqual(str_var.values, [[str_val]])

    def test_1601_insert_many(self):
        "1601 - test insert (multiple rows) with DML returning"
        self.cursor.execute("truncate table TestTempTable")
        int_values = [5, 8, 17, 24, 6]
        str_values = ["Test 5", "Test 8", "Test 17", "Test 24", "Test 6"]
        int_var = self.cursor.var(oracledb.NUMBER, arraysize=len(int_values))
        str_var = self.cursor.var(str, arraysize=len(int_values))
        self.cursor.setinputsizes(None, None, int_var, str_var)
        data = list(zip(int_values, str_values))
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var""", data)
        self.assertEqual(int_var.values, [[v] for v in int_values])
        self.assertEqual(str_var.values, [[v] for v in str_values])

    def test_1602_insert_with_small_size(self):
        "1602 - test insert with DML returning into too small a variable"
        self.cursor.execute("truncate table TestTempTable")
        int_val = 6
        str_val = "A different test string"
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str, 2)
        parameters = dict(int_val=int_val, str_val=str_val, int_var=int_var,
                          str_var=str_var)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4002:|^DPI-1037:",
                self.cursor.execute, """
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var""",
                parameters)

    def test_1603_update_single_row(self):
        "1603 - test update single row with DML returning"
        int_val = 7
        str_val = "The updated value of the string"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""",
                (int_val, "The initial value of the string"))
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        self.cursor.execute("""
                update TestTempTable set
                    StringCol1 = :str_val
                where IntCol = :int_val
                returning IntCol, StringCol1 into :int_var, :str_var""",
                int_val=int_val,
                str_val=str_val,
                int_var=int_var,
                str_var=str_var)
        self.assertEqual(int_var.values, [[int_val]])
        self.assertEqual(str_var.values, [[str_val]])

    def test_1604_update_no_rows(self):
        "1604 - test update no rows with DML returning"
        int_val = 8
        str_val = "The updated value of the string"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""",
                (int_val, "The initial value of the string"))
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        self.cursor.execute("""
                update TestTempTable set
                    StringCol1 = :str_val
                where IntCol = :int_val
                returning IntCol, StringCol1 into :int_var, :str_var""",
                int_val=int_val + 1,
                str_val=str_val,
                int_var=int_var,
                str_var=str_var)
        self.assertEqual(int_var.values, [[]])
        self.assertEqual(str_var.values, [[]])
        self.assertEqual(int_var.getvalue(), [])
        self.assertEqual(str_var.getvalue(), [])

    def test_1605_update_multiple_rows(self):
        "1605 - test update multiple rows with DML returning"
        self.cursor.execute("truncate table TestTempTable")
        for i in (8, 9, 10):
            self.cursor.execute("""
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)""",
                    (i, "The initial value of string %d" % i))
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        self.cursor.execute("""
                update TestTempTable set
                    IntCol = IntCol + 15,
                    StringCol1 = 'The final value of string ' || to_char(IntCol)
                returning IntCol, StringCol1 into :int_var, :str_var""",
                int_var=int_var,
                str_var=str_var)
        self.assertEqual(self.cursor.rowcount, 3)
        self.assertEqual(int_var.values, [[23, 24, 25]])
        expected_values = [[
            "The final value of string 8",
            "The final value of string 9",
            "The final value of string 10"
        ]]
        self.assertEqual(str_var.values, expected_values)

    def test_1606_update_multiple_rows_executemany(self):
        "1606 - test update multiple rows with DML returning (executemany)"
        data = [(i, "The initial value of string %d" % i) \
                for i in range(1, 11)]
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""", data)
        int_var = self.cursor.var(oracledb.NUMBER, arraysize=3)
        str_var = self.cursor.var(str, arraysize=3)
        self.cursor.setinputsizes(None, int_var, str_var)
        self.cursor.executemany("""
                update TestTempTable set
                    IntCol = IntCol + 25,
                    StringCol1 = 'Updated value of string ' || to_char(IntCol)
                where IntCol < :inVal
                returning IntCol, StringCol1 into :int_var, :str_var""",
                [[3], [8], [11]])
        expected_values = [
            [26, 27],
            [28, 29, 30, 31, 32],
            [33, 34, 35]
        ]
        self.assertEqual(int_var.values, expected_values)
        expected_values = [
            [
                "Updated value of string 1",
                "Updated value of string 2"
            ],
            [
                "Updated value of string 3",
                "Updated value of string 4",
                "Updated value of string 5",
                "Updated value of string 6",
                "Updated value of string 7"
            ],
            [
                "Updated value of string 8",
                "Updated value of string 9",
                "Updated value of string 10"
            ]
        ]
        self.assertEqual(str_var.values, expected_values)

    def test_1607_insert_and_return_object(self):
        "1607 - test inserting an object with DML returning"
        type_obj = self.connection.gettype("UDT_OBJECT")
        string_value = "The string that will be verified"
        obj = type_obj.newobject()
        obj.STRINGVALUE = string_value
        out_var = self.cursor.var(oracledb.DB_TYPE_OBJECT,
                                  typename="UDT_OBJECT")
        self.cursor.execute("""
                insert into TestObjects (IntCol, ObjectCol)
                values (4, :obj)returning ObjectCol into :outObj""",
                obj=obj, outObj=out_var)
        result, = out_var.getvalue()
        self.assertEqual(result.STRINGVALUE, string_value)
        self.connection.rollback()

    def test_1608_insert_and_return_rowid(self):
        "1608 - test inserting a row and returning a rowid"
        self.cursor.execute("truncate table TestTempTable")
        var = self.cursor.var(oracledb.ROWID)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (278, 'String 278')
                returning rowid into :1""", (var,))
        rowid, = var.getvalue()
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                where rowid = :1""",
                (rowid,))
        self.assertEqual(self.cursor.fetchall(), [(278, 'String 278')])

    def test_1609_insert_with_ref_cursor(self):
        "1609 - test inserting with a REF cursor and returning a rowid"
        self.cursor.execute("truncate table TestTempTable")
        var = self.cursor.var(oracledb.ROWID)
        in_cursor = self.connection.cursor()
        in_cursor.execute("""
                select StringCol
                from TestStrings
                where IntCol >= 5
                order by IntCol""")
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (187, pkg_TestRefCursors.TestInCursor(:1))
                returning rowid into :2""", (in_cursor, var))
        rowid, = var.getvalue()
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                where rowid = :1""",
                (rowid,))
        self.assertEqual(self.cursor.fetchall(),
                         [(187, 'String 7 (Modified)')])

    def test_1610_delete_returning_decreasing_rows_returned(self):
        "1610 - test delete returning decreasing number of rows"
        data = [(i, "Test String %d" % i) for i in range(1, 11)]
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""", data)
        results = []
        int_var = self.cursor.var(int)
        self.cursor.setinputsizes(None, int_var)
        for int_val in (5, 8, 10):
            self.cursor.execute("""
                    delete from TestTempTable
                    where IntCol < :1
                    returning IntCol into :2""", [int_val])
            results.append(int_var.getvalue())
        self.assertEqual(results, [[1, 2, 3, 4], [5, 6, 7], [8, 9]])

    def test_1611_delete_returning_no_rows_after_many_rows(self):
        "1611 - test delete returning no rows after returning many rows"
        data = [(i, "Test String %d" % i) for i in range(1, 11)]
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""", data)
        int_var = self.cursor.var(int)
        self.cursor.execute("""
                delete from TestTempTable
                where IntCol < :1
                returning IntCol into :2""", [5, int_var])
        self.assertEqual(int_var.getvalue(), [1, 2, 3, 4])
        self.cursor.execute(None, [4, int_var])
        self.assertEqual(int_var.getvalue(), [])

    def test_1612_insert_with_dml_returning_and_error(self):
        "1612 - test DML returning when an error occurs"
        self.cursor.execute("truncate table TestTempTable")
        int_val = 7
        str_val = "A" * 401
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        sql = """
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var"""
        parameters = dict(int_val=int_val, str_val=str_val, int_var=int_var,
                          str_var=str_var)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-12899:",
                               self.cursor.execute, sql, parameters)

    def test_1613_insert_with_dml_returning_no_input_vars(self):
        "1613 - test DML returning with no input variables, multiple iters"
        self.cursor.execute("truncate table TestTempTable")
        sql = """
                insert into TestTempTable (IntCol)
                values ((select count(*) + 1 from TestTempTable))
                returning IntCol into :1"""
        var = self.cursor.var(int)
        self.cursor.execute(sql, [var])
        self.assertEqual(var.getvalue(), [1])
        self.cursor.execute(sql, [var])
        self.assertEqual(var.getvalue(), [2])

    def test_1614_parse_quoted_returning_bind(self):
        "1614 - test DML returning with a quoted bind name"
        sql = '''
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :"_val1" , :"VaL_2"'''
        self.cursor.parse(sql)
        expected_bind_names = ['INT_VAL', 'STR_VAL', '_val1', 'VaL_2']
        self.assertEqual(self.cursor.bindnames(), expected_bind_names)

    def test_1615_parse_invalid_returning_bind(self):
        "1615 - test DML returning with an invalid bind name"
        sql = '''
                insert into TestTempTable (IntCol)
                values (:int_val)
                returning IntCol, StringCol1 into :ROWID'''
        self.assertRaisesRegex(oracledb.DatabaseError,
                               "^ORA-01745:", self.cursor.parse, sql)

    def test_1616_parse_non_ascii_returning_bind(self):
        "1616 - test DML returning with a non-ascii bind name"
        sql = '''
                insert into TestTempTable (IntCol)
                values (:int_val)
                returning IntCol, StringCol1 into :méil'''
        self.cursor.prepare(sql)
        self.assertEqual(self.cursor.bindnames(), ["INT_VAL", "MÉIL"])

    def test_1617_dml_returning_with_input_bind_vars(self):
        "1617 - test DML returning with input bind variable data"
        self.cursor.execute("truncate table TestTempTable")
        out_var = self.cursor.var(int)
        self.cursor.execute("""
                insert into TestTempTable (IntCol)
                values (:int_val)
                returning IntCol + :add_val into :out_val""",
                int_val=5,
                add_val=18,
                out_val=out_var)
        self.connection.commit()
        self.assertEqual(out_var.getvalue(), [23])

    def test_1618_dml_returning_with_lob_and_outconverter(self):
        "1618 - test DML returning with LOBs and an output converter"
        self.cursor.execute("truncate table TestCLOBs")
        out_var = self.cursor.var(oracledb.DB_TYPE_CLOB,
                                  outconverter=lambda value: value.read())
        lob_value = "A short CLOB - 1618"
        self.cursor.execute("""
                insert into TestCLOBs
                (IntCol, ClobCol)
                values (1, :in_val)
                returning CLOBCol into :out_val""",
                in_val=lob_value,
                out_val=out_var)
        self.connection.commit()
        self.assertEqual(out_var.getvalue(), [lob_value])

    def test_1619_dml_returning_with_clob_converted_to_long(self):
        "1619 - test DML returning with CLOB converted to LONG"
        self.cursor.execute("truncate table TestCLOBs")
        out_var = self.cursor.var(oracledb.DB_TYPE_LONG)
        lob_value = "A short CLOB - 1619"
        self.cursor.execute("""
                insert into TestCLOBs
                (IntCol, ClobCol)
                values (1, :in_val)
                returning CLOBCol into :out_val""",
                in_val=lob_value,
                out_val=out_var)
        self.connection.commit()
        self.assertEqual(out_var.getvalue(), [lob_value])

    def test_1620_dml_returning_with_index_organized_table(self):
        "1620 - test dml returning with an index organized table"
        self.cursor.execute("truncate table TestUniversalRowids")
        rowid_var = self.cursor.var(oracledb.ROWID)
        data = (1, "ABC", datetime.datetime(2017, 4, 11), rowid_var)
        sql = "insert into TestUniversalRowids values (:1, :2, :3)\n" + \
              "returning rowid into :4"
        self.cursor.execute(sql, data)
        rowid_value, = rowid_var.getvalue()
        self.cursor.execute("""
                select * from TestUniversalRowids where rowid = :1""",
                [rowid_value])
        row, = self.cursor.fetchall()
        self.assertEqual(data[:3], row)

    def test_1621_plsql_returning_rowids_with_index_organized_table(self):
        "1621 - test plsql returning rowids with index organized table"
        self.cursor.execute("truncate table TestUniversalRowids")
        rowid_var = self.cursor.var(oracledb.ROWID)
        data = (1, "ABC", datetime.datetime(2017, 4, 11), rowid_var)
        self.cursor.execute("""
                begin
                insert into TestUniversalRowids values (:1, :2, :3)
                returning rowid into :4; end;""", data)
        rowid_value = rowid_var.values[0]
        self.cursor.execute("""
                select * from TestUniversalRowids where rowid = :1""",
                [rowid_value])
        row, = self.cursor.fetchall()
        self.assertEqual(data[:3], row)

    def test_1622_parse_returning_clause_without_spaces(self):
        "1622 - parse DML returning with no spaces"
        self.cursor.execute("truncate table TestTempTable")
        sql = 'insert into TestTempTable (IntCol) values (:in_val)' + \
              'returning(IntCol)into :out_val'
        out_val = self.cursor.var(int, arraysize=5)
        self.cursor.execute(sql, in_val=25, out_val=out_val)
        self.assertEqual(out_val.getvalue(), [25])

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# constants.pxi
#
# Cython file defining constants used by the thin implementation classes
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

# packet types
DEF TNS_PACKET_TYPE_CONNECT = 1
DEF TNS_PACKET_TYPE_ACCEPT = 2
DEF TNS_PACKET_TYPE_REFUSE = 4
DEF TNS_PACKET_TYPE_DATA = 6
DEF TNS_PACKET_TYPE_RESEND = 11
DEF TNS_PACKET_TYPE_MARKER = 12
DEF TNS_PACKET_TYPE_CONTROL = 14
DEF TNS_PACKET_TYPE_REDIRECT = 5

# packet flags
DEF TNS_PACKET_FLAG_REDIRECT = 0x04
DEF TNS_PACKET_FLAG_TLS_RENEG = 0x08

# data flags
DEF TNS_DATA_FLAGS_EOF = 0x0040

# marker types
DEF TNS_MARKER_TYPE_BREAK = 1
DEF TNS_MARKER_TYPE_RESET = 2
DEF TNS_MARKER_TYPE_INTERRUPT = 3

# charset forms
DEF TNS_CS_IMPLICIT = 1
DEF TNS_CS_NCHAR = 2

# data types
DEF TNS_DATA_TYPE_VARCHAR = 1
DEF TNS_DATA_TYPE_NUMBER = 2
DEF TNS_DATA_TYPE_BINARY_INTEGER = 3
DEF TNS_DATA_TYPE_FLOAT = 4
DEF TNS_DATA_TYPE_STR = 5
DEF TNS_DATA_TYPE_VNU = 6
DEF TNS_DATA_TYPE_PDN = 7
DEF TNS_DATA_TYPE_LONG = 8
DEF TNS_DATA_TYPE_VCS = 9
DEF TNS_DATA_TYPE_TIDDEF = 10
DEF TNS_DATA_TYPE_ROWID = 11
DEF TNS_DATA_TYPE_DATE = 12
DEF TNS_DATA_TYPE_VBI = 15
DEF TNS_DATA_TYPE_RAW = 23
DEF TNS_DATA_TYPE_LONG_RAW = 24
DEF TNS_DATA_TYPE_UB2 = 25
DEF TNS_DATA_TYPE_UB4 = 26
DEF TNS_DATA_TYPE_SB1 = 27
DEF TNS_DATA_TYPE_SB2 = 28
DEF TNS_DATA_TYPE_SB4 = 29
DEF TNS_DATA_TYPE_SWORD = 30
DEF TNS_DATA_TYPE_UWORD = 31
DEF TNS_DATA_TYPE_PTRB = 32
DEF TNS_DATA_TYPE_PTRW = 33
DEF TNS_DATA_TYPE_OER8 = 34 + 256
DEF TNS_DATA_TYPE_FUN = 35 + 256
DEF TNS_DATA_TYPE_AUA = 36 + 256
DEF TNS_DATA_TYPE_RXH7 = 37 + 256
DEF TNS_DATA_TYPE_NA6 = 38 + 256
DEF TNS_DATA_TYPE_OAC = 39
DEF TNS_DATA_TYPE_AMS = 40
DEF TNS_DATA_TYPE_BRN = 41
DEF TNS_DATA_TYPE_BRP = 42 + 256
DEF TNS_DATA_TYPE_BRV = 43 + 256
DEF TNS_DATA_TYPE_KVA = 44 + 256
DEF TNS_DATA_TYPE_CLS = 45 + 256
DEF TNS_DATA_TYPE_CUI = 46 + 256
DEF TNS_DATA_TYPE_DFN = 47 + 256
DEF TNS_DATA_TYPE_DQR = 48 + 256
DEF TNS_DATA_TYPE_DSC = 49 + 256
DEF TNS_DATA_TYPE_EXE = 50 + 256
DEF TNS_DATA_TYPE_FCH = 51 + 256
DEF TNS_DATA_TYPE_GBV = 52 + 256
DEF TNS_DATA_TYPE_GEM = 53 + 256
DEF TNS_DATA_TYPE_GIV = 54 + 256
DEF TNS_DATA_TYPE_OKG = 55 + 256
DEF TNS_DATA_TYPE_HMI = 56 + 256
DEF TNS_DATA_TYPE_INO = 57 + 256
DEF TNS_DATA_TYPE_LNF = 59 + 256
DEF TNS_DATA_TYPE_ONT = 60 + 256
DEF TNS_DATA_TYPE_OPE = 61 + 256
DEF TNS_DATA_TYPE_OSQ = 62 + 256
DEF TNS_DATA_TYPE_SFE = 63 + 256
DEF TNS_DATA_TYPE_SPF = 64 + 256
DEF TNS_DATA_TYPE_VSN = 65 + 256
DEF TNS_DATA_TYPE_UD7 = 66 + 256
DEF TNS_DATA_TYPE_DSA = 67 + 256
DEF TNS_DATA_TYPE_UIN = 68
DEF TNS_DATA_TYPE_PIN = 71 + 256
DEF TNS_DATA_TYPE_PFN = 72 + 256
DEF TNS_DATA_TYPE_PPT = 73 + 256
DEF TNS_DATA_TYPE_STO = 75 + 256
DEF TNS_DATA_TYPE_ARC = 77 + 256
DEF TNS_DATA_TYPE_MRS = 78 + 256
DEF TNS_DATA_TYPE_MRT = 79 + 256
DEF TNS_DATA_TYPE_MRG = 80 + 256
DEF TNS_DATA_TYPE_MRR = 81 + 256
DEF TNS_DATA_TYPE_MRC = 82 + 256
DEF TNS_DATA_TYPE_VER = 83 + 256
DEF TNS_DATA_TYPE_LON2 = 84 + 256
DEF TNS_DATA_TYPE_INO2 = 85 + 256
DEF TNS_DATA_TYPE_ALL = 86 + 256
DEF TNS_DATA_TYPE_UDB = 87 + 256
DEF TNS_DATA_TYPE_AQI = 88 + 256
DEF TNS_DATA_TYPE_ULB = 89 + 256
DEF TNS_DATA_TYPE_ULD = 90 + 256
DEF TNS_DATA_TYPE_SLS = 91
DEF TNS_DATA_TYPE_SID = 92 + 256
DEF TNS_DATA_TYPE_NA7 = 93 + 256
DEF TNS_DATA_TYPE_LVC = 94
DEF TNS_DATA_TYPE_LVB = 95
DEF TNS_DATA_TYPE_CHAR = 96
DEF TNS_DATA_TYPE_AVC = 97
DEF TNS_DATA_TYPE_AL7 = 98 + 256
DEF TNS_DATA_TYPE_K2RPC = 99 + 256
DEF TNS_DATA_TYPE_BINARY_FLOAT = 100
DEF TNS_DATA_TYPE_BINARY_DOUBLE = 101
DEF TNS_DATA_TYPE_CURSOR = 102
DEF TNS_DATA_TYPE_RDD = 104
DEF TNS_DATA_TYPE_XDP = 103 + 256
DEF TNS_DATA_TYPE_OSL = 106
DEF TNS_DATA_TYPE_OKO8 = 107 + 256
DEF TNS_DATA_TYPE_EXT_NAMED = 108
DEF TNS_DATA_TYPE_INT_NAMED = 109
DEF TNS_DATA_TYPE_EXT_REF = 110
DEF TNS_DATA_TYPE_INT_REF = 111
DEF TNS_DATA_TYPE_CLOB = 112
DEF TNS_DATA_TYPE_BLOB = 113
DEF TNS_DATA_TYPE_BFILE = 114
DEF TNS_DATA_TYPE_CFILE = 115
DEF TNS_DATA_TYPE_RSET = 116
DEF TNS_DATA_TYPE_CWD = 117
DEF TNS_DATA_TYPE_JSON = 119
DEF TNS_DATA_TYPE_NEW_OAC = 120
DEF TNS_DATA_TYPE_UD12 = 124 + 256
DEF TNS_DATA_TYPE_AL8 = 125 + 256
DEF TNS_DATA_TYPE_LFOP = 126 + 256
DEF TNS_DATA_TYPE_FCRT = 127 + 256
DEF TNS_DATA_TYPE_DNY = 128 + 256
DEF TNS_DATA_TYPE_OPR = 129 + 256
DEF TNS_DATA_TYPE_PLS = 130 + 256
DEF TNS_DATA_TYPE_XID = 131 + 256
DEF TNS_DATA_TYPE_TXN = 132 + 256
DEF TNS_DATA_TYPE_DCB = 133 + 256
DEF TNS_DATA_TYPE_CCA = 134 + 256
DEF TNS_DATA_TYPE_WRN = 135 + 256
DEF TNS_DATA_TYPE_TLH = 137 + 256
DEF TNS_DATA_TYPE_TOH = 138 + 256
DEF TNS_DATA_TYPE_FOI = 139 + 256
DEF TNS_DATA_TYPE_SID2 = 140 + 256
DEF TNS_DATA_TYPE_TCH = 141 + 256
DEF TNS_DATA_TYPE_PII = 142 + 256
DEF TNS_DATA_TYPE_PFI = 143 + 256
DEF TNS_DATA_TYPE_PPU = 144 + 256
DEF TNS_DATA_TYPE_PTE = 145 + 256
DEF TNS_DATA_TYPE_CLV = 146
DEF TNS_DATA_TYPE_RXH8 = 148 + 256
DEF TNS_DATA_TYPE_N12 = 149 + 256
DEF TNS_DATA_TYPE_AUTH = 150 + 256
DEF TNS_DATA_TYPE_KVAL = 151 + 256
DEF TNS_DATA_TYPE_DTR = 152
DEF TNS_DATA_TYPE_DUN = 153
DEF TNS_DATA_TYPE_DOP = 154
DEF TNS_DATA_TYPE_VST = 155
DEF TNS_DATA_TYPE_ODT = 156
DEF TNS_DATA_TYPE_FGI = 157 + 256
DEF TNS_DATA_TYPE_DSY = 158 + 256
DEF TNS_DATA_TYPE_DSYR8 = 159 + 256
DEF TNS_DATA_TYPE_DSYH8 = 160 + 256
DEF TNS_DATA_TYPE_DSYL = 161 + 256
DEF TNS_DATA_TYPE_DSYT8 = 162 + 256
DEF TNS_DATA_TYPE_DSYV8 = 163 + 256
DEF TNS_DATA_TYPE_DSYP = 164 + 256
DEF TNS_DATA_TYPE_DSYF = 165 + 256
DEF TNS_DATA_TYPE_DSYK = 166 + 256
DEF TNS_DATA_TYPE_DSYY = 167 + 256
DEF TNS_DATA_TYPE_DSYQ = 168 + 256
DEF TNS_DATA_TYPE_DSYC = 169 + 256
DEF TNS_DATA_TYPE_DSYA = 170 + 256
DEF TNS_DATA_TYPE_OT8 = 171 + 256
DEF TNS_DATA_TYPE_DOL = 172
DEF TNS_DATA_TYPE_DSYTY = 173 + 256
DEF TNS_DATA_TYPE_AQE = 174 + 256
DEF TNS_DATA_TYPE_KV = 175 + 256
DEF TNS_DATA_TYPE_AQD = 176 + 256
DEF TNS_DATA_TYPE_AQ8 = 177 + 256
DEF TNS_DATA_TYPE_TIME = 178
DEF TNS_DATA_TYPE_TIME_TZ = 179
DEF TNS_DATA_TYPE_TIMESTAMP = 180
DEF TNS_DATA_TYPE_TIMESTAMP_TZ = 181
DEF TNS_DATA_TYPE_INTERVAL_YM = 182
DEF TNS_DATA_TYPE_INTERVAL_DS = 183
DEF TNS_DATA_TYPE_EDATE = 184
DEF TNS_DATA_TYPE_ETIME = 185
DEF TNS_DATA_TYPE_ETTZ = 186
DEF TNS_DATA_TYPE_ESTAMP = 187
DEF TNS_DATA_TYPE_ESTZ = 188
DEF TNS_DATA_TYPE_EIYM = 189
DEF TNS_DATA_TYPE_EIDS = 190
DEF TNS_DATA_TYPE_RFS = 193 + 256
DEF TNS_DATA_TYPE_RXH10 = 194 + 256
DEF TNS_DATA_TYPE_DCLOB = 195
DEF TNS_DATA_TYPE_DBLOB = 196
DEF TNS_DATA_TYPE_DBFILE = 197
DEF TNS_DATA_TYPE_DJSON = 198
DEF TNS_DATA_TYPE_KPN = 198 + 256
DEF TNS_DATA_TYPE_KPDNR = 199 + 256
DEF TNS_DATA_TYPE_DSYD = 200 + 256
DEF TNS_DATA_TYPE_DSYS = 201 + 256
DEF TNS_DATA_TYPE_DSYR = 202 + 256
DEF TNS_DATA_TYPE_DSYH = 203 + 256
DEF TNS_DATA_TYPE_DSYT = 204 + 256
DEF TNS_DATA_TYPE_DSYV = 205 + 256
DEF TNS_DATA_TYPE_AQM = 206 + 256
DEF TNS_DATA_TYPE_OER11 = 207 + 256
DEF TNS_DATA_TYPE_UROWID = 208
DEF TNS_DATA_TYPE_AQL = 210 + 256
DEF TNS_DATA_TYPE_OTC = 211 + 256
DEF TNS_DATA_TYPE_KFNO = 212 + 256
DEF TNS_DATA_TYPE_KFNP = 213 + 256
DEF TNS_DATA_TYPE_KGT8 = 214 + 256
DEF TNS_DATA_TYPE_RASB4 = 215 + 256
DEF TNS_DATA_TYPE_RAUB2 = 216 + 256
DEF TNS_DATA_TYPE_RAUB1 = 217 + 256
DEF TNS_DATA_TYPE_RATXT = 218 + 256
DEF TNS_DATA_TYPE_RSSB4 = 219 + 256
DEF TNS_DATA_TYPE_RSUB2 = 220 + 256
DEF TNS_DATA_TYPE_RSUB1 = 221 + 256
DEF TNS_DATA_TYPE_RSTXT = 222 + 256
DEF TNS_DATA_TYPE_RIDL = 223 + 256
DEF TNS_DATA_TYPE_GLRDD = 224 + 256
DEF TNS_DATA_TYPE_GLRDG = 225 + 256
DEF TNS_DATA_TYPE_GLRDC = 226 + 256
DEF TNS_DATA_TYPE_OKO = 227 + 256
DEF TNS_DATA_TYPE_DPP = 228 + 256
DEF TNS_DATA_TYPE_DPLS = 229 + 256
DEF TNS_DATA_TYPE_DPMOP = 230 + 256
DEF TNS_DATA_TYPE_TIMESTAMP_LTZ = 231
DEF TNS_DATA_TYPE_ESITZ = 232
DEF TNS_DATA_TYPE_UB8 = 233
DEF TNS_DATA_TYPE_STAT = 234 + 256
DEF TNS_DATA_TYPE_RFX = 235 + 256
DEF TNS_DATA_TYPE_FAL = 236 + 256
DEF TNS_DATA_TYPE_CKV = 237 + 256
DEF TNS_DATA_TYPE_DRCX = 238 + 256
DEF TNS_DATA_TYPE_KGH = 239 + 256
DEF TNS_DATA_TYPE_AQO = 240 + 256
DEF TNS_DATA_TYPE_PNTY = 241
DEF TNS_DATA_TYPE_OKGT = 242 + 256
DEF TNS_DATA_TYPE_KPFC = 243 + 256
DEF TNS_DATA_TYPE_FE2 = 244 + 256
DEF TNS_DATA_TYPE_SPFP = 245 + 256
DEF TNS_DATA_TYPE_DPULS = 246 + 256
DEF TNS_DATA_TYPE_BOOLEAN = 252
DEF TNS_DATA_TYPE_AQA = 253 + 256
DEF TNS_DATA_TYPE_KPBF = 254 + 256
DEF TNS_DATA_TYPE_TSM = 513
DEF TNS_DATA_TYPE_MSS = 514
DEF TNS_DATA_TYPE_KPC = 516
DEF TNS_DATA_TYPE_CRS = 517
DEF TNS_DATA_TYPE_KKS = 518
DEF TNS_DATA_TYPE_KSP = 519
DEF TNS_DATA_TYPE_KSPTOP = 520
DEF TNS_DATA_TYPE_KSPVAL = 521
DEF TNS_DATA_TYPE_PSS = 522
DEF TNS_DATA_TYPE_NLS = 523
DEF TNS_DATA_TYPE_ALS = 524
DEF TNS_DATA_TYPE_KSDEVTVAL = 525
DEF TNS_DATA_TYPE_KSDEVTTOP = 526
DEF TNS_DATA_TYPE_KPSPP = 527
DEF TNS_DATA_TYPE_KOL = 528
DEF TNS_DATA_TYPE_LST = 529
DEF TNS_DATA_TYPE_ACX = 530
DEF TNS_DATA_TYPE_SCS = 531
DEF TNS_DATA_TYPE_RXH = 532
DEF TNS_DATA_TYPE_KPDNS = 533
DEF TNS_DATA_TYPE_KPDCN = 534
DEF TNS_DATA_TYPE_KPNNS = 535
DEF TNS_DATA_TYPE_KPNCN = 536
DEF TNS_DATA_TYPE_KPS = 537
DEF TNS_DATA_TYPE_APINF = 538
DEF TNS_DATA_TYPE_TEN = 539
DEF TNS_DATA_TYPE_XSSCS = 540
DEF TNS_DATA_TYPE_XSSSO = 541
DEF TNS_DATA_TYPE_XSSAO = 542
DEF TNS_DATA_TYPE_KSRPC = 543
DEF TNS_DATA_TYPE_KVL = 560
DEF TNS_DATA_TYPE_SESSGET = 563
DEF TNS_DATA_TYPE_SESSREL = 564
DEF TNS_DATA_TYPE_XSSDEF = 565
DEF TNS_DATA_TYPE_PDQCINV = 572
DEF TNS_DATA_TYPE_PDQIDC = 573
DEF TNS_DATA_TYPE_KPDQCSTA = 574
DEF TNS_DATA_TYPE_KPRS = 575
DEF TNS_DATA_TYPE_KPDQIDC = 576
DEF TNS_DATA_TYPE_RTSTRM = 578
DEF TNS_DATA_TYPE_SESSRET = 579
DEF TNS_DATA_TYPE_SCN6 = 580
DEF TNS_DATA_TYPE_KECPA = 581
DEF TNS_DATA_TYPE_KECPP = 582
DEF TNS_DATA_TYPE_SXA = 583
DEF TNS_DATA_TYPE_KVARR = 584
DEF TNS_DATA_TYPE_KPNGN = 585
DEF TNS_DATA_TYPE_XSNSOP = 590
DEF TNS_DATA_TYPE_XSATTR = 591
DEF TNS_DATA_TYPE_XSNS = 592
DEF TNS_DATA_TYPE_TXT = 593
DEF TNS_DATA_TYPE_XSSESSNS = 594
DEF TNS_DATA_TYPE_XSATTOP = 595
DEF TNS_DATA_TYPE_XSCREOP = 596
DEF TNS_DATA_TYPE_XSDETOP = 597
DEF TNS_DATA_TYPE_XSDESOP = 598
DEF TNS_DATA_TYPE_XSSETSP = 599
DEF TNS_DATA_TYPE_XSSIDP = 600
DEF TNS_DATA_TYPE_XSPRIN = 601
DEF TNS_DATA_TYPE_XSKVL = 602
DEF TNS_DATA_TYPE_XSSSDEF2 = 603
DEF TNS_DATA_TYPE_XSNSOP2 = 604
DEF TNS_DATA_TYPE_XSNS2 = 605
DEF TNS_DATA_TYPE_IMPLRES = 611
DEF TNS_DATA_TYPE_OER = 612
DEF TNS_DATA_TYPE_UB1ARRAY = 613
DEF TNS_DATA_TYPE_SESSSTATE = 614
DEF TNS_DATA_TYPE_AC_REPLAY = 615
DEF TNS_DATA_TYPE_AC_CONT = 616
DEF TNS_DATA_TYPE_KPDNREQ = 622
DEF TNS_DATA_TYPE_KPDNRNF = 623
DEF TNS_DATA_TYPE_KPNGNC = 624
DEF TNS_DATA_TYPE_KPNRI = 625
DEF TNS_DATA_TYPE_AQENQ = 626
DEF TNS_DATA_TYPE_AQDEQ = 627
DEF TNS_DATA_TYPE_AQJMS = 628
DEF TNS_DATA_TYPE_KPDNRPAY = 629
DEF TNS_DATA_TYPE_KPDNRACK = 630
DEF TNS_DATA_TYPE_KPDNRMP = 631
DEF TNS_DATA_TYPE_KPDNRDQ = 632
DEF TNS_DATA_TYPE_CHUNKINFO = 636
DEF TNS_DATA_TYPE_SCN = 637
DEF TNS_DATA_TYPE_SCN8 = 638
DEF TNS_DATA_TYPE_UDS = 639
DEF TNS_DATA_TYPE_TNP = 640

# data type representations
DEF TNS_TYPE_REP_NATIVE = 0
DEF TNS_TYPE_REP_UNIVERSAL = 1
DEF TNS_TYPE_REP_ORACLE = 10

# errors
DEF TNS_ERR_VAR_NOT_IN_SELECT_LIST = 1007
DEF TNS_ERR_INBAND_MESSAGE = 12573
DEF TNS_ERR_INVALID_SERVICE_NAME = 12514
DEF TNS_ERR_INVALID_SID = 12505
DEF TNS_ERR_NO_DATA_FOUND = 1403
DEF TNS_ERR_SESSION_SHUTDOWN = 12572
DEF TNS_ERR_ARRAY_DML_ERRORS = 24381

# message types
DEF TNS_MSG_TYPE_PROTOCOL = 1
DEF TNS_MSG_TYPE_DATA_TYPES = 2
DEF TNS_MSG_TYPE_FUNCTION = 3
DEF TNS_MSG_TYPE_ERROR = 4
DEF TNS_MSG_TYPE_ROW_HEADER = 6
DEF TNS_MSG_TYPE_ROW_DATA = 7
DEF TNS_MSG_TYPE_PARAMETER = 8
DEF TNS_MSG_TYPE_STATUS = 9
DEF TNS_MSG_TYPE_IO_VECTOR = 11
DEF TNS_MSG_TYPE_LOB_DATA = 14
DEF TNS_MSG_TYPE_WARNING = 15
DEF TNS_MSG_TYPE_DESCRIBE_INFO = 16
DEF TNS_MSG_TYPE_PIGGYBACK = 17
DEF TNS_MSG_TYPE_FLUSH_OUT_BINDS = 19
DEF TNS_MSG_TYPE_BIT_VECTOR = 21
DEF TNS_MSG_TYPE_SERVER_SIDE_PIGGYBACK = 23
DEF TNS_MSG_TYPE_ONEWAY_FN = 26
DEF TNS_MSG_TYPE_IMPLICIT_RESULTSET = 27

# parameter keyword numbers
DEF TNS_KEYWORD_NUM_CURRENT_SCHEMA = 168
DEF TNS_KEYWORD_NUM_EDITION = 172

# bind flags
DEF TNS_BIND_USE_INDICATORS = 0x0001
DEF TNS_BIND_ARRAY = 0x0040

# bind directions
DEF TNS_BIND_DIR_OUTPUT = 16
DEF TNS_BIND_DIR_INPUT = 32
DEF TNS_BIND_DIR_INPUT_OUTPUT = 48

# database object image flags
DEF TNS_OBJ_IS_VERSION_81 = 0x80
DEF TNS_OBJ_IS_DEGENERATE = 0x10
DEF TNS_OBJ_IS_COLLECTION = 0x08
DEF TNS_OBJ_NO_PREFIX_SEG = 0x04
DEF TNS_OBJ_IMAGE_VERSION = 1

# database object flags
DEF TNS_OBJ_MAX_SHORT_LENGTH = 245
DEF TNS_OBJ_ATOMIC_NULL = 253
DEF TNS_OBJ_NON_NULL_OID = 0x02
DEF TNS_OBJ_HAS_EXTENT_OID = 0x08
DEF TNS_OBJ_TOP_LEVEL = 0x01
DEF TNS_OBJ_HAS_INDEXES = 0x10

# database object collection types
DEF TNS_OBJ_PLSQL_INDEX_TABLE = 1
DEF TNS_OBJ_NESTED_TABLE = 2
DEF TNS_OBJ_VARRAY = 3

# database object TDS type codes
DEF TNS_OBJ_TDS_TYPE_CHAR = 1
DEF TNS_OBJ_TDS_TYPE_DATE = 2
DEF TNS_OBJ_TDS_TYPE_FLOAT = 5
DEF TNS_OBJ_TDS_TYPE_NUMBER = 6
DEF TNS_OBJ_TDS_TYPE_VARCHAR = 7
DEF TNS_OBJ_TDS_TYPE_BOOLEAN = 8
DEF TNS_OBJ_TDS_TYPE_RAW = 19
DEF TNS_OBJ_TDS_TYPE_TIMESTAMP = 21
DEF TNS_OBJ_TDS_TYPE_TIMESTAMP_TZ = 23
DEF TNS_OBJ_TDS_TYPE_OBJ = 27
DEF TNS_OBJ_TDS_TYPE_COLL = 28
DEF TNS_OBJ_TDS_TYPE_CLOB = 29
DEF TNS_OBJ_TDS_TYPE_BLOB = 30
DEF TNS_OBJ_TDS_TYPE_TIMESTAMP_LTZ = 33
DEF TNS_OBJ_TDS_TYPE_BINARY_FLOAT = 37
DEF TNS_OBJ_TDS_TYPE_BINARY_DOUBLE = 45

# xml type constants
DEF TNS_XML_TYPE_LOB = 0x0001
DEF TNS_XML_TYPE_STRING = 0x0004
DEF TNS_XML_TYPE_FLAG_SKIP_NEXT_4 = 0x100000

# execute options
DEF TNS_EXEC_OPTION_PARSE = 0x01
DEF TNS_EXEC_OPTION_BIND = 0x08
DEF TNS_EXEC_OPTION_DEFINE = 0x10
DEF TNS_EXEC_OPTION_EXECUTE = 0x20
DEF TNS_EXEC_OPTION_FETCH = 0x40
DEF TNS_EXEC_OPTION_COMMIT = 0x100
DEF TNS_EXEC_OPTION_COMMIT_REEXECUTE = 0x1
DEF TNS_EXEC_OPTION_PLSQL_BIND = 0x400
DEF TNS_EXEC_OPTION_DML_ROWCOUNTS = 0x4000
DEF TNS_EXEC_OPTION_NOT_PLSQL = 0x8000
DEF TNS_EXEC_OPTION_IMPLICIT_RESULTSET = 0x8000
DEF TNS_EXEC_OPTION_DESCRIBE = 0x20000
DEF TNS_EXEC_OPTION_NO_COMPRESSED_FETCH = 0x40000
DEF TNS_EXEC_OPTION_BATCH_ERRORS = 0x80000

# server side piggyback op codes
DEF TNS_SERVER_PIGGYBACK_QUERY_CACHE_INVALIDATION = 1
DEF TNS_SERVER_PIGGYBACK_OS_PID_MTS = 2
DEF TNS_SERVER_PIGGYBACK_TRACE_EVENT = 3
DEF TNS_SERVER_PIGGYBACK_SESS_RET = 4
DEF TNS_SERVER_PIGGYBACK_SYNC = 5
DEF TNS_SERVER_PIGGYBACK_LTXID = 7
DEF TNS_SERVER_PIGGYBACK_AC_REPLAY_CONTEXT = 8
DEF TNS_SERVER_PIGGYBACK_EXT_SYNC = 9

# session return constants
DEF TNS_SESSGET_SESSION_CHANGED = 4

# LOB operations
DEF TNS_LOB_OP_GET_LENGTH = 0x0001
DEF TNS_LOB_OP_READ = 0x0002
DEF TNS_LOB_OP_TRIM = 0x0020
DEF TNS_LOB_OP_WRITE = 0x0040
DEF TNS_LOB_OP_GET_CHUNK_SIZE = 0x4000
DEF TNS_LOB_OP_CREATE_TEMP = 0x0110
DEF TNS_LOB_OP_FREE_TEMP = 0x0111
DEF TNS_LOB_OP_OPEN = 0x8000
DEF TNS_LOB_OP_CLOSE = 0x10000
DEF TNS_LOB_OP_IS_OPEN = 0x11000
DEF TNS_LOB_OP_ARRAY = 0x80000

# LOB locator constants
DEF TNS_LOB_LOC_OFFSET_FLAG_1 = 4
DEF TNS_LOB_LOC_OFFSET_FLAG_3 = 6
DEF TNS_LOB_LOC_OFFSET_FLAG_4 = 7
DEF TNS_LOB_QLOCATOR_VERSION = 4

# LOB locator flags (byte 1)
DEF TNS_LOB_LOC_FLAGS_BLOB = 0x01
DEF TNS_LOB_LOC_FLAGS_VALUE_BASED = 0x20
DEF TNS_LOB_LOC_FLAGS_ABSTRACT = 0x40

# LOB locator flags (byte 2)
DEF TNS_LOB_LOC_FLAGS_INIT = 0x08

# LOB locator flags (byte 4)
DEF TNS_LOB_LOC_FLAGS_TEMP = 0x01
DEF TNS_LOB_LOC_FLAGS_VAR_LENGTH_CHARSET = 0x80

# other LOB constants
DEF TNS_LOB_OPEN_READ_WRITE = 2
DEF TNS_LOB_PREFETCH_FLAG = 0x2000000

# base JSON constants
DEF TNS_JSON_MAX_LENGTH = 32 * 1024 * 1024
DEF TNS_JSON_MAGIC_BYTE_1 = 0xff
DEF TNS_JSON_MAGIC_BYTE_2 = 0x4a        # 'J'
DEF TNS_JSON_MAGIC_BYTE_3 = 0x5a        # 'Z'
DEF TNS_JSON_VERSION = 1
DEF TNS_JSON_FLAG_HASH_ID_UINT8 = 0x0100
DEF TNS_JSON_FLAG_HASH_ID_UINT16 = 0x0200
DEF TNS_JSON_FLAG_NUM_FNAMES_UINT16 = 0x0400
DEF TNS_JSON_FLAG_FNAMES_SEG_UINT32 = 0x0800
DEF TNS_JSON_FLAG_TINY_NODES_STAT = 0x2000
DEF TNS_JSON_FLAG_TREE_SEG_UINT32 = 0x1000
DEF TNS_JSON_FLAG_REL_OFFSET_MODE = 0x01
DEF TNS_JSON_FLAG_INLINE_LEAF = 0x02
DEF TNS_JSON_FLAG_LEN_IN_PCODE = 0x04
DEF TNS_JSON_FLAG_NUM_FNAMES_UINT32 = 0x08
DEF TNS_JSON_FLAG_IS_SCALAR = 0x10

# JSON data types
DEF TNS_JSON_TYPE_NULL = 0x30
DEF TNS_JSON_TYPE_TRUE = 0x31
DEF TNS_JSON_TYPE_FALSE = 0x32
DEF TNS_JSON_TYPE_STRING_LENGTH_UINT8 = 0x33
DEF TNS_JSON_TYPE_NUMBER_LENGTH_UINT8 = 0x34
DEF TNS_JSON_TYPE_BINARY_DOUBLE = 0x36
DEF TNS_JSON_TYPE_STRING_LENGTH_UINT16 = 0x37
DEF TNS_JSON_TYPE_STRING_LENGTH_UINT32 = 0x38
DEF TNS_JSON_TYPE_TIMESTAMP = 0x39
DEF TNS_JSON_TYPE_BINARY_LENGTH_UINT16 = 0x3a
DEF TNS_JSON_TYPE_BINARY_LENGTH_UINT32 = 0x3b
DEF TNS_JSON_TYPE_DATE = 0x3c
DEF TNS_JSON_TYPE_INTERVAL_YM = 0x3d
DEF TNS_JSON_TYPE_INTERVAL_DS = 0x3e
DEF TNS_JSON_TYPE_TIMESTAMP_TZ = 0x7c
DEF TNS_JSON_TYPE_TIMESTAMP7 = 0x7d
DEF TNS_JSON_TYPE_BINARY_FLOAT = 0x7f
DEF TNS_JSON_TYPE_OBJECT = 0x84
DEF TNS_JSON_TYPE_ARRAY = 0xc0

# end-to-end metrics
DEF TNS_END_TO_END_ACTION = 0x0010
DEF TNS_END_TO_END_CLIENT_IDENTIFIER = 0x0001
DEF TNS_END_TO_END_CLIENT_INFO = 0x0100
DEF TNS_END_TO_END_DBOP = 0x0200
DEF TNS_END_TO_END_MODULE = 0x0008

# versions
DEF TNS_VERSION_DESIRED = 319
DEF TNS_VERSION_MINIMUM = 300
DEF TNS_VERSION_MIN_ACCEPTED = 315      # 12.1
DEF TNS_VERSION_MIN_LARGE_SDU = 315
DEF TNS_VERSION_MIN_OOB_CHECK = 318

# control packet types
DEF TNS_CONTROL_TYPE_INBAND_NOTIFICATION = 8
DEF TNS_CONTROL_TYPE_RESET_OOB = 9

# connect flags
DEF TNS_GSO_DONT_CARE = 0x0001
DEF TNS_GSO_CAN_RECV_ATTENTION = 0x0400
DEF TNS_NSI_DISABLE_NA = 0x04
DEF TNS_NSI_SUPPORT_SECURITY_RENEG = 0x80

# other connection constants
DEF TNS_PROTOCOL_CHARACTERISTICS = 0x4f98
DEF TNS_CHECK_OOB = 0x01

# TTC functions
DEF TNS_FUNC_AUTH_PHASE_ONE = 118
DEF TNS_FUNC_AUTH_PHASE_TWO = 115
DEF TNS_FUNC_CLOSE_CURSORS = 105
DEF TNS_FUNC_COMMIT = 14
DEF TNS_FUNC_EXECUTE = 94
DEF TNS_FUNC_FETCH = 5
DEF TNS_FUNC_LOB_OP = 96
DEF TNS_FUNC_LOGOFF = 9
DEF TNS_FUNC_PING = 147
DEF TNS_FUNC_ROLLBACK = 15
DEF TNS_FUNC_SET_END_TO_END_ATTR = 135
DEF TNS_FUNC_REEXECUTE = 4
DEF TNS_FUNC_REEXECUTE_AND_FETCH = 78
DEF TNS_FUNC_SESSION_GET = 162
DEF TNS_FUNC_SESSION_RELEASE = 163
DEF TNS_FUNC_SET_SCHEMA = 152

# TTC authentication modes
DEF TNS_AUTH_MODE_LOGON = 0x00000001
DEF TNS_AUTH_MODE_CHANGE_PASSWORD = 0x00000002
DEF TNS_AUTH_MODE_SYSDBA = 0x00000020
DEF TNS_AUTH_MODE_SYSOPER = 0x00000040
DEF TNS_AUTH_MODE_PRELIM = 0x00000080
DEF TNS_AUTH_MODE_WITH_PASSWORD = 0x00000100
DEF TNS_AUTH_MODE_SYSASM = 0x00400000
DEF TNS_AUTH_MODE_SYSBKP = 0x01000000
DEF TNS_AUTH_MODE_SYSDGD = 0x02000000
DEF TNS_AUTH_MODE_SYSKMT = 0x04000000
DEF TNS_AUTH_MODE_SYSRAC = 0x08000000
DEF TNS_AUTH_MODE_IAM_TOKEN = 0x20000000

# character sets and encodings
DEF TNS_CHARSET_UTF8 = 873
DEF TNS_CHARSET_UTF16 = 2000
DEF TNS_ENCODING_UTF8 = "UTF-8"
DEF TNS_ENCODING_UTF16 = "UTF-16BE"
DEF TNS_ENCODING_MULTI_BYTE = 0x01
DEF TNS_ENCODING_CONV_LENGTH = 0x02

# compile time capability indices
DEF TNS_CCAP_SQL_VERSION = 0
DEF TNS_CCAP_LOGON_TYPES = 4
DEF TNS_CCAP_FIELD_VERSION = 7
DEF TNS_CCAP_SERVER_DEFINE_CONV = 8
DEF TNS_CCAP_TTC1 = 15
DEF TNS_CCAP_OCI1 = 16
DEF TNS_CCAP_TDS_VERSION = 17
DEF TNS_CCAP_RPC_VERSION = 18
DEF TNS_CCAP_RPC_SIG = 19
DEF TNS_CCAP_DBF_VERSION = 21
DEF TNS_CCAP_LOB = 23
DEF TNS_CCAP_TTC2 = 26
DEF TNS_CCAP_UB2_DTY = 27
DEF TNS_CCAP_OCI2 = 31
DEF TNS_CCAP_CLIENT_FN = 34
DEF TNS_CCAP_TTC3 = 37
DEF TNS_CCAP_TTC4 = 40
DEF TNS_CCAP_LOB2 = 42
DEF TNS_CCAP_MAX = 45

# compile time capability values
DEF TNS_CCAP_SQL_VERSION_MAX = 6
DEF TNS_CCAP_FIELD_VERSION_11_2 = 6
DEF TNS_CCAP_FIELD_VERSION_12_1 = 7
DEF TNS_CCAP_FIELD_VERSION_12_2 = 8
DEF TNS_CCAP_FIELD_VERSION_12_2_EXT1 = 9
DEF TNS_CCAP_FIELD_VERSION_18_1 = 10
DEF TNS_CCAP_FIELD_VERSION_18_1_EXT_1 = 11
DEF TNS_CCAP_FIELD_VERSION_19_1 = 12
DEF TNS_CCAP_FIELD_VERSION_19_1_EXT_1 = 13
DEF TNS_CCAP_FIELD_VERSION_20_1 = 14
DEF TNS_CCAP_FIELD_VERSION_20_1_EXT_1 = 15
DEF TNS_CCAP_FIELD_VERSION_21_1 = 16
DEF TNS_CCAP_FIELD_VERSION_23_1 = 17
DEF TNS_CCAP_FIELD_VERSION_23_1_EXT_1 = 18
DEF TNS_CCAP_FIELD_VERSION_MAX = 18
DEF TNS_CCAP_O5LOGON = 8
DEF TNS_CCAP_O5LOGON_NP = 2
DEF TNS_CCAP_O7LOGON = 32
DEF TNS_CCAP_O8LOGON_LONG_IDENTIFIER = 64
DEF TNS_CCAP_O9LOGON_LONG_PASSWORD = 0x80
DEF TNS_CCAP_END_OF_CALL_STATUS = 0x01
DEF TNS_CCAP_IND_RCD = 0x08
DEF TNS_CCAP_FAST_BVEC = 0x20
DEF TNS_CCAP_FAST_SESSION_PROPAGATE = 0x10
DEF TNS_CCAP_APP_CTX_PIGGYBACK = 0x80
DEF TNS_CCAP_TDS_VERSION_MAX = 3
DEF TNS_CCAP_RPC_VERSION_MAX = 7
DEF TNS_CCAP_RPC_SIG_VALUE = 3
DEF TNS_CCAP_DBF_VERSION_MAX = 1
DEF TNS_CCAP_IMPLICIT_RESULTS = 0x10
DEF TNS_CCAP_BIG_CHUNK_CLR = 0x20
DEF TNS_CCAP_KEEP_OUT_ORDER = 0x80
DEF TNS_CCAP_LOB_UB8_SIZE = 0x01
DEF TNS_CCAP_LOB_ENCS = 0x02
DEF TNS_CCAP_LOB_PREFETCH_DATA = 0x04
DEF TNS_CCAP_LOB_TEMP_SIZE = 0x08
DEF TNS_CCAP_LOB_PREFETCH_LENGTH = 0x40
DEF TNS_CCAP_LOB_12C = 0x80
DEF TNS_CCAP_LOB2_QUASI = 0x01
DEF TNS_CCAP_LOB2_2GB_PREFETCH = 0x04
DEF TNS_CCAP_DRCP = 0x10
DEF TNS_CCAP_ZLNP = 0x04
DEF TNS_CCAP_INBAND_NOTIFICATION = 0x04
DEF TNS_CCAP_CLIENT_FN_MAX = 12

# runtime capability indices
DEF TNS_RCAP_COMPAT = 0
DEF TNS_RCAP_TTC = 6
DEF TNS_RCAP_MAX = 7

# runtime capability values
DEF TNS_RCAP_COMPAT_81 = 2
DEF TNS_RCAP_TTC_ZERO_COPY = 0x01
DEF TNS_RCAP_TTC_32K = 0x04

# verifier types
DEF TNS_VERIFIER_TYPE_11G_1 = 0xb152
DEF TNS_VERIFIER_TYPE_11G_2 = 0x1b25
DEF TNS_VERIFIER_TYPE_12C = 0x4815

# other constants
DEF TNS_MAX_SHORT_LENGTH = 252
DEF TNS_ESCAPE_CHAR = 253
DEF TNS_LONG_LENGTH_INDICATOR = 254
DEF TNS_NULL_LENGTH_INDICATOR = 255
DEF TNS_MAX_ROWID_LENGTH = 18
DEF TNS_DURATION_MID = 0x80000000
DEF TNS_DURATION_OFFSET = 60
DEF TNS_DURATION_SESSION = 10
DEF TNS_MAX_LONG_LENGTH = 0x7fffffff
DEF TNS_SDU = 8192
DEF TNS_TDU = 65535
DEF TNS_MAX_CURSORS_TO_CLOSE = 500
DEF TNS_TXN_IN_PROGRESS = 0x00000002
DEF TNS_MAX_CONNECT_DATA = 230
DEF TNS_CHUNK_SIZE = 32767
DEF TNS_MAX_UROWID_LENGTH = 5267
DEF TNS_HAS_REGION_ID = 0x80

# base 64 encoding alphabet
DEF TNS_BASE64_ALPHABET = \
        b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
cdef bytearray TNS_BASE64_ALPHABET_ARRAY = \
        bytearray(TNS_BASE64_ALPHABET)
cdef bytes TNS_EXTENT_OID = bytes.fromhex('00000000000000000000000000010001')

# purity types
DEF PURITY_DEFAULT = 0
DEF PURITY_NEW = 1
DEF PURITY_SELF = 2

# timezone offsets
DEF TZ_HOUR_OFFSET = 20
DEF TZ_MINUTE_OFFSET = 60

# drcp release mode
DEF DRCP_DEAUTHENTICATE = 0x00000002



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject.pyx
#
# Cython file defining the thick DbObjectType, DbObjectAttr and DbObject
# implementation classes (embedded in thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThickDbObjectImpl(BaseDbObjectImpl):
    cdef:
        dpiObject *_handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiObject_release(self._handle)

    cdef int _convert_from_python(self, object value, DbType dbtype,
                                  ThickDbObjectTypeImpl objtype, dpiData *data,
                                  StringBuffer buf):
        """
        Internal method for converting a value from Python to the value
        required by ODPI-C.
        """
        if value is None:
            data.isNull = 1
        else:
            data.isNull = 0
            _convert_from_python(value, dbtype, objtype, &data.value, buf)

    def append_checked(self, object value):
        """
        Internal method for appending a value to a collection object.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            ThickDbObjectTypeImpl objtype
            dpiData data
        objtype = <ThickDbObjectTypeImpl> self.type
        self._convert_from_python(value, objtype.element_dbtype,
                                  objtype.element_objtype, &data, buf)
        if dpiObject_appendElement(self._handle,
                                   objtype._element_native_type_num,
                                   &data) < 0:
            _raise_from_odpi()

    def copy(self):
        """
        Internal method for creating a copy of an object.
        """
        cdef ThickDbObjectImpl copied_impl
        copied_impl = ThickDbObjectImpl.__new__(ThickDbObjectImpl)
        if dpiObject_copy(self._handle, &copied_impl._handle) < 0:
            _raise_from_odpi()
        copied_impl.type = self.type
        return copied_impl

    def delete_by_index(self, int32_t index):
        """
        Internal method for deleting an entry from a collection that is indexed
        by integers.
        """
        if dpiObject_deleteElementByIndex(self._handle, index) < 0:
            _raise_from_odpi()

    def exists_by_index(self, int32_t index):
        """
        Internal method for determining if an entry exists in a collection that
        is indexed by integers.
        """
        cdef bint exists
        if dpiObject_getElementExistsByIndex(self._handle, index, &exists) < 0:
            _raise_from_odpi()
        return exists

    def get_attr_value(self, ThickDbObjectAttrImpl attr):
        """
        Internal method for getting an attribute value.
        """
        cdef:
            char number_as_string_buffer[200]
            ThickDbObjectTypeImpl type_impl
            dpiData data
        if attr._native_type_num == DPI_NATIVE_TYPE_BYTES \
                and attr.dbtype.num == DPI_ORACLE_TYPE_NUMBER:
            data.value.asBytes.ptr = number_as_string_buffer
            data.value.asBytes.length = sizeof(number_as_string_buffer)
            data.value.asBytes.encoding = NULL
        if dpiObject_getAttributeValue(self._handle, attr._handle,
                                       attr._native_type_num, &data) < 0:
            _raise_from_odpi()
        if data.isNull:
            return None
        type_impl = self.type
        return _convert_to_python(type_impl._conn_impl, attr.dbtype,
                                  attr.objtype, &data.value)

    def get_element_by_index(self, int32_t index):
        """
        Internal method for getting an entry from a collection that is indexed
        by integers.
        """
        cdef:
            char number_as_string_buffer[200]
            ThickDbObjectTypeImpl objtype
            dpiData data
        objtype = self.type
        if objtype._element_native_type_num == DPI_NATIVE_TYPE_BYTES \
                and objtype.element_dbtype.num == DPI_ORACLE_TYPE_NUMBER:
            data.value.asBytes.ptr = number_as_string_buffer
            data.value.asBytes.length = sizeof(number_as_string_buffer)
            data.value.asBytes.encoding = NULL
        if dpiObject_getElementValueByIndex(self._handle, index,
                                            objtype._element_native_type_num,
                                            &data) < 0:
            _raise_from_odpi()
        if data.isNull:
            return None
        return _convert_to_python(objtype._conn_impl, objtype.element_dbtype,
                                  objtype.element_objtype, &data.value)

    def get_first_index(self):
        """
        Internal method for getting the first index from a collection that is
        indexed by integers.
        """
        cdef:
            int32_t index
            bint exists
        if dpiObject_getFirstIndex(self._handle, &index, &exists) < 0:
            _raise_from_odpi()
        if exists:
            return index

    def get_last_index(self):
        """
        Internal method for getting the last index from a collection that is
        indexed by integers.
        """
        cdef:
            int32_t index
            bint exists
        if dpiObject_getLastIndex(self._handle, &index, &exists) < 0:
            _raise_from_odpi()
        if exists:
            return index

    def get_next_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        cdef:
            int32_t next_index
            bint exists
        if dpiObject_getNextIndex(self._handle, index, &next_index,
                                  &exists) < 0:
            _raise_from_odpi()
        if exists:
            return next_index

    def get_prev_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        cdef:
            int32_t prev_index
            bint exists
        if dpiObject_getPrevIndex(self._handle, index, &prev_index,
                                  &exists) < 0:
            _raise_from_odpi()
        if exists:
            return prev_index

    def get_size(self):
        """
        Internal method for getting the size of a collection.
        """
        cdef int32_t size
        if dpiObject_getSize(self._handle, &size) < 0:
            _raise_from_odpi()
        return size

    def set_attr_value_checked(self, ThickDbObjectAttrImpl attr, object value):
        """
        Internal method for setting an attribute value.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            uint32_t native_type_num
            dpiData data
        self._convert_from_python(value, attr.dbtype, attr.objtype, &data,
                                  buf)
        native_type_num = attr._native_type_num
        if native_type_num == DPI_NATIVE_TYPE_LOB \
                and not isinstance(value, PY_TYPE_LOB):
            native_type_num = DPI_NATIVE_TYPE_BYTES
        if dpiObject_setAttributeValue(self._handle, attr._handle,
                                       native_type_num, &data) < 0:
            _raise_from_odpi()

    def set_element_by_index_checked(self, int32_t index, object value):
        """
        Internal method for setting an entry in a collection that is indexed by
        integers.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            ThickDbObjectTypeImpl objtype
            uint32_t native_type_num
            dpiData data
        objtype = self.type
        self._convert_from_python(value, objtype.element_dbtype,
                                  objtype.element_objtype, &data, buf)
        native_type_num = objtype._element_native_type_num
        if native_type_num == DPI_NATIVE_TYPE_LOB \
                and not isinstance(value, PY_TYPE_LOB):
            native_typeNum = DPI_NATIVE_TYPE_BYTES
        if dpiObject_setElementValueByIndex(self._handle, index,
                                            native_type_num, &data) < 0:
            _raise_from_odpi()

    def trim(self, int32_t num_to_trim):
        """
        Internal method for trimming a number of entries from a collection.
        """
        if dpiObject_trim(self._handle, num_to_trim) < 0:
            _raise_from_odpi()


cdef class ThickDbObjectAttrImpl(BaseDbObjectAttrImpl):
    cdef:
        dpiObjectAttr *_handle
        uint32_t _native_type_num

    def __dealloc__(self):
        if self._handle != NULL:
            dpiObjectAttr_release(self._handle)

    @staticmethod
    cdef ThickDbObjectAttrImpl _from_handle(ThickConnImpl conn_impl,
                                            dpiObjectAttr *handle):
        """
        Create a new DbObjectAttr implementation object given an ODPI-C handle.
        """
        cdef:
            ThickDbObjectAttrImpl impl
            dpiObjectType *typ_handle
            dpiObjectAttrInfo info
        impl = ThickDbObjectAttrImpl.__new__(ThickDbObjectAttrImpl)
        impl._handle = handle
        if dpiObjectAttr_getInfo(handle, &info) < 0:
            _raise_from_odpi()
        impl.name = info.name[:info.nameLength].decode()
        impl.dbtype = DbType._from_num(info.typeInfo.oracleTypeNum)
        if info.typeInfo.objectType:
            typ_handle = info.typeInfo.objectType
            impl.objtype = ThickDbObjectTypeImpl._from_handle(conn_impl,
                                                              typ_handle)
        impl._native_type_num = _get_native_type_num(impl.dbtype)
        return impl


cdef class ThickDbObjectTypeImpl(BaseDbObjectTypeImpl):
    cdef:
        dpiObjectType *_handle
        uint32_t _element_native_type_num

    def __dealloc__(self):
        if self._handle != NULL:
            dpiObjectType_release(self._handle)

    @staticmethod
    cdef ThickDbObjectTypeImpl _from_handle(ThickConnImpl conn_impl,
                                            dpiObjectType *handle):
        """
        Create a new DbObjectType implementation object given an ODPI-C handle.
        """
        cdef:
            dpiObjectAttr **attributes = NULL
            ThickDbObjectTypeImpl impl, temp
            ThickDbObjectAttrImpl attr_impl
            dpiObjectTypeInfo info
            ssize_t num_bytes
            DbType dbtype
            uint16_t i
            object typ
        impl = ThickDbObjectTypeImpl.__new__(ThickDbObjectTypeImpl)
        if dpiObjectType_addRef(handle) < 0:
            _raise_from_odpi()
        impl._conn_impl = conn_impl
        impl._handle = handle
        if dpiObjectType_getInfo(impl._handle, &info) < 0:
            _raise_from_odpi()
        impl.schema = info.schema[:info.schemaLength].decode()
        impl.name = info.name[:info.nameLength].decode()
        if info.packageNameLength > 0:
            impl.package_name = \
                    info.packageName[:info.packageNameLength].decode()
        impl.is_collection = info.isCollection
        if impl.is_collection:
            dbtype = DbType._from_num(info.elementTypeInfo.oracleTypeNum)
            impl.element_dbtype = dbtype
            if info.elementTypeInfo.objectType != NULL:
                handle = info.elementTypeInfo.objectType
                temp = ThickDbObjectTypeImpl._from_handle(conn_impl, handle)
                impl.element_objtype = temp
            impl._element_native_type_num = \
                    _get_native_type_num(impl.element_dbtype)
        impl.attrs_by_name = {}
        impl.attrs = [None] * info.numAttributes
        try:
            num_bytes = info.numAttributes * sizeof(dpiObjectAttr*)
            attributes = <dpiObjectAttr**> cpython.PyMem_Malloc(num_bytes)
            if dpiObjectType_getAttributes(impl._handle, info.numAttributes,
                                           attributes) < 0:
                _raise_from_odpi()
            for i in range(info.numAttributes):
                attr_impl = ThickDbObjectAttrImpl._from_handle(conn_impl,
                                                               attributes[i])
                impl.attrs[i] = attr_impl
                impl.attrs_by_name[attr_impl.name] = attr_impl
        finally:
            if attributes != NULL:
                cpython.PyMem_Free(attributes)
        return impl

    def create_new_object(self):
        """
        Internal method for creating a new object.
        """
        cdef ThickDbObjectImpl obj_impl
        obj_impl = ThickDbObjectImpl.__new__(ThickDbObjectImpl)
        obj_impl.type = self
        if dpiObjectType_createObject(self._handle, &obj_impl._handle) < 0:
            _raise_from_odpi()
        return obj_impl
